import os
from typing import Generator

import asyncpg
import pytest

from ddtrace import Pin
from ddtrace import tracer
from ddtrace.contrib.asyncpg import patch
from ddtrace.contrib.asyncpg import unpatch
from ddtrace.contrib.trace_utils import iswrapped
from tests.contrib.config import POSTGRES_CONFIG


@pytest.fixture(autouse=True)
def patch_asyncpg():
    # type: () -> Generator[None, None, None]
    patch()
    yield
    unpatch()


@pytest.fixture
async def patched_conn():
    # type: () -> Generator[asyncpg.Connection, None, None]
    conn = await asyncpg.connect(
        host=POSTGRES_CONFIG["host"],
        port=POSTGRES_CONFIG["port"],
        user=POSTGRES_CONFIG["user"],
        database=POSTGRES_CONFIG["dbname"],
        password=POSTGRES_CONFIG["password"],
    )
    yield conn
    await conn.close()


@pytest.mark.asyncio
async def test_connect(snapshot_context):
    with snapshot_context():
        conn = await asyncpg.connect(
            host=POSTGRES_CONFIG["host"],
            port=POSTGRES_CONFIG["port"],
            user=POSTGRES_CONFIG["user"],
            database=POSTGRES_CONFIG["dbname"],
            password=POSTGRES_CONFIG["password"],
        )
        await conn.close()

    # Using dsn should result in the same trace
    with snapshot_context():
        conn = await asyncpg.connect(
            dsn="postgresql://%s:%s@%s:%s/%s"
            % (
                POSTGRES_CONFIG["user"],
                POSTGRES_CONFIG["password"],
                POSTGRES_CONFIG["host"],
                POSTGRES_CONFIG["port"],
                POSTGRES_CONFIG["dbname"],
            )
        )
        await conn.close()


@pytest.mark.asyncio
@pytest.mark.snapshot(
    ignores=["meta.error.stack", "meta.error.msg", "meta.error.type"]
)  # stack is noisy between releases
async def test_bad_connect():
    with pytest.raises(OSError):
        await asyncpg.connect(
            host="localhost",
            port=POSTGRES_CONFIG["port"] + 1,
        )


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_connection_methods(patched_conn):
    status = await patched_conn.execute(
        """
        CREATE TEMP TABLE test (id serial PRIMARY KEY, name varchar(12) NOT NULL UNIQUE);
    """
    )
    assert status == "CREATE TABLE"

    status = await patched_conn.executemany(
        """
        INSERT INTO test (name) VALUES ($1), ($2), ($3);
    """,
        [["val1", "val2", "val3"]],
    )
    assert status is None

    records = await patched_conn.fetch("SELECT * FROM test;")
    assert len(records) == 3

    val = await patched_conn.fetchval("SELECT * FROM test LIMIT 1;", column=1)
    assert val == "val1"

    row = await patched_conn.fetchrow("SELECT * FROM test LIMIT 1;")
    assert len(row) == 2
    assert row["name"] == "val1"


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_select(patched_conn):
    ret = await patched_conn.fetchval("SELECT 1")
    assert ret == 1


@pytest.mark.asyncio
@pytest.mark.snapshot(ignores=["meta.error.stack"])  # stack is noisy between releases
async def test_bad_query(patched_conn):
    with pytest.raises(asyncpg.exceptions.PostgresSyntaxError):
        await patched_conn.execute("malformed; query;dfaskjfd")


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_cursor(patched_conn):
    await patched_conn.execute(
        """
        CREATE TEMP TABLE test (id serial PRIMARY KEY, name varchar(12) NOT NULL UNIQUE);
    """
    )
    await patched_conn.execute(
        """
        INSERT INTO test (name) VALUES ($1), ($2);
        """,
        "value1",
        "value2",
    )

    records = []
    async with patched_conn.transaction():
        async for r in patched_conn.cursor("SELECT * FROM test;"):
            records.append(r["name"])

    assert records == ["value1", "value2"]


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_cursor_manual(patched_conn):
    async with patched_conn.transaction():
        cur = await patched_conn.cursor("SELECT generate_series(0, 100)")
        await cur.forward(10)
        await cur.fetchrow()
        await cur.fetch(5)


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_service_override_pin(patched_conn):
    Pin.override(patched_conn, service="custom-svc")
    await patched_conn.execute("SELECT 1")


@pytest.mark.asyncio
@pytest.mark.snapshot
async def test_parenting(patched_conn):
    with tracer.trace("parent"):
        await patched_conn.execute("SELECT 1")

    with tracer.trace("parent2"):
        c = patched_conn.execute("SELECT 1")
    await c


@pytest.mark.snapshot(async_mode=False)
def test_configure_service_name_env(ddtrace_run_python_code_in_subprocess):
    code = """
import asyncio
import sys
import asyncpg
from tests.contrib.config import POSTGRES_CONFIG

async def test():
    conn = await asyncpg.connect(
        host=POSTGRES_CONFIG["host"],
        port=POSTGRES_CONFIG["port"],
        user=POSTGRES_CONFIG["user"],
        database=POSTGRES_CONFIG["dbname"],
        password=POSTGRES_CONFIG["password"],
    )
    await conn.execute("SELECT 1")
    await conn.close()

if sys.version_info >= (3, 7, 0):
    asyncio.run(test())
else:
    asyncio.get_event_loop().run_until_complete(test())
    """
    env = os.environ.copy()
    env["DD_ASYNCPG_SERVICE"] = "global-service-name"
    out, err, status, pid = ddtrace_run_python_code_in_subprocess(code, env=env)
    assert status == 0, err
    assert err == b""


def test_patch_unpatch_asyncpg():
    assert iswrapped(asyncpg.connect)
    assert iswrapped(asyncpg.protocol.Protocol.execute)
    assert iswrapped(asyncpg.protocol.Protocol.bind_execute)
    assert iswrapped(asyncpg.protocol.Protocol.query)
    assert iswrapped(asyncpg.protocol.Protocol.bind_execute_many)
    unpatch()
    assert not iswrapped(asyncpg.connect)
    assert not iswrapped(asyncpg.protocol.Protocol.execute)
    assert not iswrapped(asyncpg.protocol.Protocol.bind_execute)
    assert not iswrapped(asyncpg.protocol.Protocol.query)
    assert not iswrapped(asyncpg.protocol.Protocol.bind_execute_many)



from time import sleep

import pytest

from ddtrace.debugging._config import config
from ddtrace.debugging._probe.model import LineProbe
from ddtrace.debugging._probe.model import Probe
from ddtrace.debugging._probe.remoteconfig import ProbePollerEvent
from ddtrace.debugging._probe.remoteconfig import ProbeRCAdapter
from ddtrace.debugging._probe.remoteconfig import _filter_by_env_and_version
from ddtrace.internal.compat import PY2
from tests.utils import override_global_config


class MockConfig(object):
    def __init__(self, *args, **kwargs):
        self.probes = {}

    @_filter_by_env_and_version
    def get_probes(self, _):
        return list(self.probes.values())

    def add_probes(self, probes):
        for probe in probes:
            self.probes[probe.probe_id] = probe

    def remove_probes(self, *probe_ids):
        for probe_id in probe_ids:
            try:
                del self.probes[probe_id]
            except KeyError:
                pass


@pytest.fixture
def mock_config():
    import ddtrace.debugging._probe.remoteconfig as rc

    original_get_probes = rc.get_probes
    mock_config = MockConfig()
    try:
        rc.get_probes = mock_config.get_probes

        yield mock_config
    finally:
        rc.get_probes = original_get_probes


@pytest.mark.parametrize(
    "env,version,expected",
    [
        (None, None, set(["probe4"])),
        (None, "dev", set(["probe3", "probe4"])),
        ("prod", None, set(["probe2", "probe4"])),
        ("prod", "dev", set(["probe1", "probe2", "probe3", "probe4"])),
    ],
)
def test_poller_env_version(env, version, expected, mock_config):
    probes = []

    def cb(e, ps):
        probes.extend(ps)

    with override_global_config(dict(env=env, version=version)):
        mock_config.add_probes(
            [
                LineProbe(
                    probe_id="probe1",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                    tags={"env": "prod", "version": "dev"},
                ),
                LineProbe(
                    probe_id="probe2",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                    tags={"env": "prod"},
                ),
                LineProbe(
                    probe_id="probe3",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                    tags={"version": "dev"},
                ),
                LineProbe(
                    probe_id="probe4",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                ),
            ]
        )

        ProbeRCAdapter(cb)(None, {})

        assert set(_.probe_id for _ in probes) == expected


@pytest.mark.xfail(PY2, reason="occasionally fails on Python 2")
def test_poller_events(mock_config):
    events = set()

    def cb(e, ps):
        events.add((e, frozenset([p.probe_id if isinstance(p, Probe) else p for p in ps])))

    mock_config.add_probes(
        [
            LineProbe(
                probe_id="probe1",
                source_file="tests/debugger/submod/stuff.py",
                line=36,
                condition=None,
            ),
            LineProbe(
                probe_id="probe2",
                source_file="tests/debugger/submod/stuff.py",
                line=36,
                condition=None,
            ),
            LineProbe(
                probe_id="probe3",
                source_file="tests/debugger/submod/stuff.py",
                line=36,
                condition=None,
            ),
            LineProbe(
                probe_id="probe4",
                source_file="tests/debugger/submod/stuff.py",
                line=36,
                condition=None,
            ),
        ]
    )

    old_interval = config.diagnostics_interval
    config.diagnostics_interval = 0.5
    try:
        adapter = ProbeRCAdapter(cb)

        adapter(None, {})
        mock_config.remove_probes("probe1", "probe2")
        mock_config.add_probes(
            [
                # Modified
                LineProbe(
                    probe_id="probe2",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                    active=False,
                ),
                # New
                LineProbe(
                    probe_id="probe5",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                    active=False,
                ),
            ]
        )
        adapter(None, {})

        # Wait to allow the next call to the adapter to generate a status event
        sleep(0.5)
        adapter(None, {})

        assert events == {
            (ProbePollerEvent.NEW_PROBES, frozenset(["probe4", "probe1", "probe2", "probe3"])),
            (ProbePollerEvent.DELETED_PROBES, frozenset(["probe1"])),
            (ProbePollerEvent.MODIFIED_PROBES, frozenset(["probe2"])),
            (ProbePollerEvent.NEW_PROBES, frozenset(["probe5"])),
            (ProbePollerEvent.STATUS_UPDATE, frozenset(["probe4", "probe2", "probe3", "probe5"])),
        }
    finally:
        config.diagnostics_interval = old_interval



import os
from typing import Optional
from typing import Tuple

from .._hooks import Hooks
from ..internal.utils.attrdict import AttrDict
from ..internal.utils.formats import asbool
from .http import HttpConfig


class IntegrationConfig(AttrDict):
    """
    Integration specific configuration object.

    This is what you will get when you do::

        from ddtrace import config

        # This is an `IntegrationConfig`
        config.flask

        # `IntegrationConfig` supports both attribute and item accessors
        config.flask['service_name'] = 'my-service-name'
        config.flask.service_name = 'my-service-name'
    """

    def __init__(self, global_config, name, *args, **kwargs):
        """
        :param global_config:
        :type global_config: Config
        :param args:
        :param kwargs:
        """
        super(IntegrationConfig, self).__init__(*args, **kwargs)

        # Set internal properties for this `IntegrationConfig`
        # DEV: By-pass the `__setattr__` overrides from `AttrDict` to set real properties
        object.__setattr__(self, "global_config", global_config)
        object.__setattr__(self, "integration_name", name)
        object.__setattr__(self, "hooks", Hooks())
        object.__setattr__(self, "http", HttpConfig())

        analytics_enabled, analytics_sample_rate = self._get_analytics_settings()
        self.setdefault("analytics_enabled", analytics_enabled)
        self.setdefault("analytics_sample_rate", float(analytics_sample_rate))
        service = os.getenv(
            "DD_%s_SERVICE" % name.upper(),
            default=os.getenv(
                "DD_%s_SERVICE_NAME" % name.upper(),
                default=None,
            ),
        )
        self.setdefault("service", service)
        # TODO[v1.0]: this is required for backwards compatibility since some
        # integrations use service_name instead of service. These should be
        # unified.
        self.setdefault("service_name", service)

    def _get_analytics_settings(self):
        # type: () -> Tuple[Optional[bool], float]
        # Set default analytics configuration, default is disabled
        # DEV: Default to `None` which means do not set this key
        # Inject environment variables for integration
        _ = os.getenv(
            "DD_TRACE_%s_ANALYTICS_ENABLED" % self.integration_name.upper(),
            os.getenv("DD_%s_ANALYTICS_ENABLED" % self.integration_name.upper()),
        )
        analytics_enabled = asbool(_) if _ is not None else None

        analytics_sample_rate = float(
            os.getenv(
                "DD_TRACE_%s_ANALYTICS_SAMPLE_RATE" % self.integration_name.upper(),
                os.getenv("DD_%s_ANALYTICS_SAMPLE_RATE" % self.integration_name.upper(), default=1.0),
            )
        )

        return analytics_enabled, analytics_sample_rate

    @property
    def trace_query_string(self):
        if self.http.trace_query_string is not None:
            return self.http.trace_query_string
        return self.global_config.http.trace_query_string

    @property
    def is_header_tracing_configured(self):
        # type: (...) -> bool
        """Returns whether header tracing is enabled for this integration.

        Will return true if traced headers are configured for this integration
        or if they are configured globally.
        """
        return self.http.is_header_tracing_configured or self.global_config.http.is_header_tracing_configured

    def header_is_traced(self, header_name):
        # type: (str) -> bool
        """
        Returns whether or not the current header should be traced.
        :param header_name: the header name
        :type header_name: str
        :rtype: bool
        """
        return self._header_tag_name(header_name) is not None

    def _header_tag_name(self, header_name):
        # type: (str) -> Optional[str]
        tag_name = self.http._header_tag_name(header_name)
        if tag_name is None:
            return self.global_config._header_tag_name(header_name)
        return tag_name

    def _is_analytics_enabled(self, use_global_config):
        # DEV: analytics flag can be None which should not be taken as
        # enabled when global flag is disabled
        if use_global_config and self.global_config.analytics_enabled:
            return self.analytics_enabled is not False
        else:
            return self.analytics_enabled is True

    def get_analytics_sample_rate(self, use_global_config=False):
        """
        Returns analytics sample rate but only when integration-specific
        analytics configuration is enabled with optional override with global
        configuration
        """
        if self._is_analytics_enabled(use_global_config):
            analytics_sample_rate = getattr(self, "analytics_sample_rate", None)
            # return True if attribute is None or attribute not found
            if analytics_sample_rate is None:
                return True
            # otherwise return rate
            return analytics_sample_rate

        # Use `None` as a way to say that it was not defined,
        #   `False` would mean `0` which is a different thing
        return None

    def __repr__(self):
        cls = self.__class__
        keys = ", ".join(self.keys())
        return "{}.{}({})".format(cls.__module__, cls.__name__, keys)



"""
This module defines Tornado settings that are shared between
integration modules.
"""
CONFIG_KEY = "datadog_trace"
REQUEST_SPAN_KEY = "__datadog_request_span"
FUTURE_SPAN_KEY = "__datadog_future_span"



from ddtrace.internal.logger import get_logger
from ddtrace.settings.dynamic_instrumentation import DynamicInstrumentationConfig


log = get_logger(__name__)

config = DynamicInstrumentationConfig()

log.debug("Dynamic instrumentation configuration: %r", config.__dict__)



import dogpile
import pytest

from ddtrace import Pin
from ddtrace.contrib.dogpile_cache.patch import patch
from ddtrace.contrib.dogpile_cache.patch import unpatch
from tests.utils import DummyTracer
from tests.utils import TracerSpanContainer
from tests.utils import assert_is_measured


@pytest.fixture
def tracer():
    return DummyTracer()


@pytest.fixture
def test_spans(tracer):
    container = TracerSpanContainer(tracer)
    yield container
    container.reset()


@pytest.fixture
def region(tracer):
    patch()
    # Setup a simple dogpile cache region for testing.
    # The backend is trivial so we can use memory to simplify test setup.
    test_region = dogpile.cache.make_region(name="TestRegion", key_mangler=lambda x: x)
    test_region.configure("dogpile.cache.memory")
    Pin.override(dogpile.cache, tracer=tracer)
    return test_region


@pytest.fixture(autouse=True)
def cleanup():
    yield
    unpatch()


@pytest.fixture
def single_cache(region):
    @region.cache_on_arguments()
    def fn(x):
        return x * 2

    return fn


@pytest.fixture
def multi_cache(region):
    @region.cache_multi_on_arguments()
    def fn(*x):
        return [i * 2 for i in x]

    return fn


def test_doesnt_trace_with_no_pin(tracer, single_cache, multi_cache, test_spans):
    # No pin is set
    unpatch()

    assert single_cache(1) == 2
    assert test_spans.pop_traces() == []

    assert multi_cache(2, 3) == [4, 6]
    assert test_spans.pop_traces() == []


def test_doesnt_trace_with_disabled_pin(tracer, single_cache, multi_cache, test_spans):
    tracer.enabled = False

    assert single_cache(1) == 2
    assert test_spans.pop_traces() == []

    assert multi_cache(2, 3) == [4, 6]
    assert test_spans.pop_traces() == []


def test_traces_get_or_create(tracer, single_cache, test_spans):
    assert single_cache(1) == 2
    traces = test_spans.pop_traces()
    assert len(traces) == 1
    spans = traces[0]
    assert len(spans) == 1
    span = spans[0]

    assert_is_measured(span)
    assert span.name == "dogpile.cache"
    assert span.span_type == "cache"
    assert span.resource == "get_or_create"
    assert span.get_tag("key") == "tests.contrib.dogpile_cache.test_tracing:fn|1"
    assert span.get_tag("hit") == "False"
    assert span.get_tag("expired") == "True"
    assert span.get_tag("backend") == "MemoryBackend"
    assert span.get_tag("region") == "TestRegion"

    # Now the results should be cached.
    assert single_cache(1) == 2
    traces = test_spans.pop_traces()
    assert len(traces) == 1
    spans = traces[0]
    assert len(spans) == 1
    span = spans[0]

    assert_is_measured(span)
    assert span.name == "dogpile.cache"
    assert span.span_type == "cache"
    assert span.resource == "get_or_create"
    assert span.get_tag("key") == "tests.contrib.dogpile_cache.test_tracing:fn|1"
    assert span.get_tag("hit") == "True"
    assert span.get_tag("expired") == "False"
    assert span.get_tag("backend") == "MemoryBackend"
    assert span.get_tag("region") == "TestRegion"


def test_traces_get_or_create_multi(tracer, multi_cache, test_spans):
    assert multi_cache(2, 3) == [4, 6]
    traces = test_spans.pop_traces()
    assert len(traces) == 1
    spans = traces[0]
    assert len(spans) == 1
    span = spans[0]

    assert_is_measured(span)
    assert span.name == "dogpile.cache"
    assert span.span_type == "cache"
    assert span.get_tag("keys") == (
        "['tests.contrib.dogpile_cache.test_tracing:fn|2', " + "'tests.contrib.dogpile_cache.test_tracing:fn|3']"
    )
    assert span.get_tag("hit") == "False"
    assert span.get_tag("expired") == "True"
    assert span.get_tag("backend") == "MemoryBackend"
    assert span.get_tag("region") == "TestRegion"

    # Partial hit
    assert multi_cache(2, 4) == [4, 8]
    traces = test_spans.pop_traces()
    assert len(traces) == 1
    spans = traces[0]
    assert len(spans) == 1
    span = spans[0]
    assert_is_measured(span)
    assert span.name == "dogpile.cache"
    assert span.span_type == "cache"
    assert span.get_tag("keys") == (
        "['tests.contrib.dogpile_cache.test_tracing:fn|2', " + "'tests.contrib.dogpile_cache.test_tracing:fn|4']"
    )
    assert span.get_tag("hit") == "False"
    assert span.get_tag("expired") == "True"
    assert span.get_tag("backend") == "MemoryBackend"
    assert span.get_tag("region") == "TestRegion"

    # Full hit
    assert multi_cache(2, 4) == [4, 8]
    traces = test_spans.pop_traces()
    assert len(traces) == 1
    spans = traces[0]
    assert len(spans) == 1
    span = spans[0]
    assert_is_measured(span)
    assert span.name == "dogpile.cache"
    assert span.span_type == "cache"
    assert span.get_tag("keys") == (
        "['tests.contrib.dogpile_cache.test_tracing:fn|2', " + "'tests.contrib.dogpile_cache.test_tracing:fn|4']"
    )
    assert span.get_tag("hit") == "True"
    assert span.get_tag("expired") == "False"
    assert span.get_tag("backend") == "MemoryBackend"
    assert span.get_tag("region") == "TestRegion"


class TestInnerFunctionCalls(object):
    def single_cache(self, x):
        return x * 2

    def multi_cache(self, *x):
        return [i * 2 for i in x]

    def test_calls_inner_functions_correctly(self, region, mocker):
        """This ensures the get_or_create behavior of dogpile is not altered."""
        spy_single_cache = mocker.spy(self, "single_cache")
        spy_multi_cache = mocker.spy(self, "multi_cache")

        single_cache = region.cache_on_arguments()(self.single_cache)
        multi_cache = region.cache_multi_on_arguments()(self.multi_cache)

        assert 2 == single_cache(1)
        spy_single_cache.assert_called_once_with(1)

        # It's now cached - shouldn't need to call the inner function.
        spy_single_cache.reset_mock()
        assert 2 == single_cache(1)
        assert spy_single_cache.call_count == 0

        assert [6, 8] == multi_cache(3, 4)
        spy_multi_cache.assert_called_once_with(3, 4)

        # Partial hit. Only the "new" key should be passed to the inner function.
        spy_multi_cache.reset_mock()
        assert [6, 10] == multi_cache(3, 5)
        spy_multi_cache.assert_called_once_with(5)

        # Full hit. No call to inner function.
        spy_multi_cache.reset_mock()
        assert [6, 10] == multi_cache(3, 5)
        assert spy_single_cache.call_count == 0


def test_get_or_create_kwarg_only(region):
    """
    When get_or_create is called with only kwargs
        The arguments should be handled correctly
    """
    assert region.get_or_create(key="key", creator=lambda: 3) == 3
    assert region.get_or_create_multi(keys="keys", creator=lambda *args: [1, 2])



from ddtrace.opentracer import Tracer


def init_tracer(service_name, dd_tracer, scope_manager=None):
    """A method that emulates what a user of OpenTracing would call to
    initialize a Datadog opentracer.

    It accepts a Datadog tracer that should be the same one used for testing.
    """
    ot_tracer = Tracer(service_name, dd_tracer=dd_tracer, scope_manager=scope_manager)
    return ot_tracer



"""
Instrumentation Telemetry API.
This is normally started automatically by ``ddtrace-run`` when the
``DD_INSTRUMENTATION_TELEMETRY_ENABLED`` variable is set.
To start the service manually, invoke the ``enable`` method::
    from ddtrace.internal.telemetry import telemetry_writer
    telemetry_writer.enable()
"""
from .writer import TelemetryWriter


telemetry_writer = TelemetryWriter()

__all__ = ["telemetry_writer"]



import grpc


GRPC_PIN_MODULE_SERVER = grpc.Server
GRPC_PIN_MODULE_CLIENT = grpc.Channel
GRPC_METHOD_PATH_KEY = "grpc.method.path"
GRPC_METHOD_PACKAGE_KEY = "grpc.method.package"
GRPC_METHOD_SERVICE_KEY = "grpc.method.service"
GRPC_METHOD_NAME_KEY = "grpc.method.name"
GRPC_METHOD_KIND_KEY = "grpc.method.kind"
GRPC_STATUS_CODE_KEY = "grpc.status.code"
GRPC_REQUEST_METADATA_PREFIX_KEY = "grpc.request.metadata."
GRPC_RESPONSE_METADATA_PREFIX_KEY = "grpc.response.metadata."
GRPC_HOST_KEY = "grpc.host"
GRPC_PORT_KEY = "grpc.port"
GRPC_SPAN_KIND_KEY = "span.kind"
GRPC_SPAN_KIND_VALUE_CLIENT = "client"
GRPC_SPAN_KIND_VALUE_SERVER = "server"
GRPC_METHOD_KIND_UNARY = "unary"
GRPC_METHOD_KIND_CLIENT_STREAMING = "client_streaming"
GRPC_METHOD_KIND_SERVER_STREAMING = "server_streaming"
GRPC_METHOD_KIND_BIDI_STREAMING = "bidi_streaming"
GRPC_SERVICE_SERVER = "grpc-server"
GRPC_AIO_SERVICE_SERVER = "grpc-aio-server"
GRPC_SERVICE_CLIENT = "grpc-client"
GRPC_AIO_SERVICE_CLIENT = "grpc-aio-client"



import asyncio
import os
import random

from sanic import Sanic
from sanic.response import json

from ddtrace import tracer
from tests.webclient import PingFilter


tracer.configure(
    settings={
        "FILTERS": [PingFilter()],
    }
)


app = Sanic("test_sanic_server")


# Depending on the version of sanic the application can be run in a child process.
# This can alter the name of the function and the default span name. Setting the span name to
# `random_sleep` ensures the snapshot tests produce consistent spans across sanic versions.
@tracer.wrap("random_sleep")
async def random_sleep():
    await asyncio.sleep(random.random() * 0.1)


@app.route("/hello")
async def hello(request):
    await random_sleep()
    return json({"hello": "world"})


@app.route("/internal_error")
async def internal_error(request):
    1 / 0


@app.route("/shutdown-tracer")
async def shutdown_tracer(request):
    tracer.shutdown()
    return json({"success": True})


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=int(os.environ["SANIC_PORT"]), access_log=False)



from io import BytesIO

import flask

from ddtrace import Pin
from ddtrace.contrib.flask import unpatch
from ddtrace.contrib.flask.patch import flask_version
from ddtrace.internal.compat import StringIO

from . import BaseFlaskTestCase


class FlaskHelpersTestCase(BaseFlaskTestCase):
    def test_patch(self):
        """
        When we patch Flask
            Then ``flask.jsonify`` is patched
            Then ``flask.send_file`` is patched
        """
        # DEV: We call `patch` in `setUp`
        self.assert_is_wrapped(flask.jsonify)
        self.assert_is_wrapped(flask.send_file)

    def test_unpatch(self):
        """
        When we unpatch Flask
            Then ``flask.jsonify`` is unpatched
            Then ``flask.send_file`` is unpatched
        """
        unpatch()
        self.assert_is_not_wrapped(flask.jsonify)
        self.assert_is_not_wrapped(flask.send_file)

    def test_jsonify(self):
        """
        When we call a patched ``flask.jsonify``
            We create a span as expected
        """
        # DEV: `jsonify` requires a active app and request contexts
        with self.app.app_context():
            with self.app.test_request_context("/"):
                response = flask.jsonify(dict(key="value"))
                self.assertTrue(isinstance(response, flask.Response))
                self.assertEqual(response.status_code, 200)

        # 1 span for `jsonify`
        # 1 span for tearing down the app context we created
        # 1 span for tearing down the request context we created
        spans = self.get_spans()
        self.assertEqual(len(spans), 3)

        self.assertIsNone(spans[0].service)
        self.assertEqual(spans[0].name, "flask.jsonify")
        self.assertEqual(spans[0].resource, "flask.jsonify")
        assert set(spans[0].get_tags().keys()) == {"runtime-id", "_dd.p.dm"}

        self.assertEqual(spans[1].name, "flask.do_teardown_request")
        self.assertEqual(spans[2].name, "flask.do_teardown_appcontext")

    def test_jsonify_pin_disabled(self):
        """
        When we call a patched ``flask.jsonify``
            When the ``flask.Flask`` ``Pin`` is disabled
                We do not create a span
        """
        # Disable the pin on the app
        pin = Pin.get_from(self.app)
        pin.tracer.enabled = False

        # DEV: `jsonify` requires a active app and request contexts
        with self.app.app_context():
            with self.app.test_request_context("/"):
                response = flask.jsonify(dict(key="value"))
                self.assertTrue(isinstance(response, flask.Response))
                self.assertEqual(response.status_code, 200)

        self.assertEqual(len(self.get_spans()), 0)

    def test_send_file(self):
        """
        When calling a patched ``flask.send_file``
            We create the expected spans
        """
        if flask_version >= (2, 0, 0):
            fp = BytesIO(b"static file")
        else:
            fp = StringIO("static file")

        with self.app.app_context():
            with self.app.test_request_context("/"):
                # DEV: Flask >= (0, 12, 0) tries to infer mimetype, so set explicitly
                response = flask.send_file(fp, mimetype="text/plain")
                self.assertTrue(isinstance(response, flask.Response))
                self.assertEqual(response.status_code, 200)

        # 1 for `send_file`
        # 1 for tearing down the request context we created
        # 1 for tearing down the app context we created
        spans = self.get_spans()
        self.assertEqual(len(spans), 3)

        self.assertEqual(spans[0].service, "flask")
        self.assertEqual(spans[0].name, "flask.send_file")
        self.assertEqual(spans[0].resource, "flask.send_file")
        assert set(spans[0].get_tags().keys()) == {"runtime-id", "_dd.p.dm"}

        self.assertEqual(spans[1].name, "flask.do_teardown_request")
        self.assertEqual(spans[2].name, "flask.do_teardown_appcontext")

    def test_send_file_pin_disabled(self):
        """
        When calling a patched ``flask.send_file``
            When the app's ``Pin`` has been disabled
                We do not create any spans
        """
        pin = Pin.get_from(self.app)
        pin.tracer.enabled = False

        if flask_version >= (2, 0, 0):
            fp = BytesIO(b"static file")
        else:
            fp = StringIO("static file")

        with self.app.app_context():
            with self.app.test_request_context("/"):
                # DEV: Flask >= (0, 12, 0) tries to infer mimetype, so set explicitly
                response = flask.send_file(fp, mimetype="text/plain")
                self.assertTrue(isinstance(response, flask.Response))
                self.assertEqual(response.status_code, 200)

        self.assertEqual(len(self.get_spans()), 0)



import pytest

import ddtrace
from ddtrace.profiling import Profiler


@pytest.fixture
def tracer(monkeypatch):
    monkeypatch.setenv("DD_TRACE_STARTUP_LOGS", "0")
    return ddtrace.Tracer()


@pytest.fixture
def profiler(monkeypatch):
    monkeypatch.setenv("DD_PROFILING_API_TIMEOUT", "0.1")
    p = Profiler()
    p.start()
    try:
        yield p
    finally:
        p.stop()



from contextlib import contextmanager
import os
import subprocess

import bm.utils as utils
import requests
import tenacity


SERVER_URL = "http://0.0.0.0:8000/"


def _get_response():
    r = requests.get(SERVER_URL)
    r.raise_for_status()


def _post_response():
    HEADERS = {
        "SERVER_PORT": "8000",
        "REMOTE_ADDR": "127.0.0.1",
        "CONTENT_TYPE": "application/json",
        "HTTP_HOST": "localhost:8000",
        "HTTP_ACCEPT": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,"
        "image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
        "HTTP_SEC_FETCH_DEST": "document",
        "HTTP_ACCEPT_ENCODING": "gzip, deflate, br",
        "HTTP_ACCEPT_LANGUAGE": "en-US,en;q=0.9",
    }
    r = requests.post(SERVER_URL + "post-view", data=utils.EXAMPLE_POST_DATA, headers=HEADERS)
    r.raise_for_status()


@tenacity.retry(
    wait=tenacity.wait_fixed(1),
    stop=tenacity.stop_after_attempt(30),
)
def _wait():
    _get_response()


@contextmanager
def server(scenario):
    env = {
        "PERF_TRACER_ENABLED": str(scenario.tracer_enabled),
        "PERF_PROFILER_ENABLED": str(scenario.profiler_enabled),
        "PERF_APPSEC_ENABLED": str(scenario.appsec_enabled),
    }
    # copy over current environ
    env.update(os.environ)
    cmd = ["gunicorn", "-c", "gunicorn.conf.py"]
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
        env=env,
    )
    # make sure process has been started
    assert proc.poll() is None
    try:
        _wait()
        if scenario.post_request:
            response = _post_response
        else:
            response = _get_response
        yield response
    finally:
        proc.terminate()
        proc.wait()



import asyncio
import functools
from typing import Callable
from typing import Tuple
from typing import Union

import grpc
from grpc import aio
from grpc.aio._typing import RequestIterableType
from grpc.aio._typing import RequestType
from grpc.aio._typing import ResponseIterableType
from grpc.aio._typing import ResponseType

from .. import trace_utils
from ... import Pin
from ... import Span
from ... import config
from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import ERROR_MSG
from ...constants import ERROR_TYPE
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...internal.compat import to_unicode
from ...propagation.http import HTTPPropagator
from ..grpc import constants
from ..grpc import utils


def create_aio_client_interceptors(pin, host, port):
    # type: (Pin, str, int) -> Tuple[aio.ClientInterceptor, ...]
    return (
        _UnaryUnaryClientInterceptor(pin, host, port),
        _UnaryStreamClientInterceptor(pin, host, port),
        _StreamUnaryClientInterceptor(pin, host, port),
        _StreamStreamClientInterceptor(pin, host, port),
    )


def _done_callback(span, code, details):
    # type: (Span, grpc.StatusCode, str) -> Callable[[aio.Call], None]
    def func(call):
        # type: (aio.Call) -> None
        try:
            span._set_str_tag(constants.GRPC_STATUS_CODE_KEY, to_unicode(code))

            # Handle server-side error in unary response RPCs
            if code != grpc.StatusCode.OK:
                _handle_error(span, call, code, details)
        finally:
            span.finish()

    return func


def _handle_error(span, call, code, details):
    # type: (Span, aio.Call, grpc.StatusCode, str) -> None
    span.error = 1
    span._set_str_tag(ERROR_MSG, details)
    span._set_str_tag(ERROR_TYPE, to_unicode(code))


def _handle_rpc_error(span, rpc_error):
    # type: (Span, aio.AioRpcError) -> None
    code = to_unicode(rpc_error.code())
    span.error = 1
    span._set_str_tag(constants.GRPC_STATUS_CODE_KEY, code)
    span._set_str_tag(ERROR_MSG, rpc_error.details())
    span._set_str_tag(ERROR_TYPE, code)
    span.finish()


async def _handle_cancelled_error(call, span):
    # type: (aio.Call, Span) -> None
    code = to_unicode(await call.code())
    span.error = 1
    span._set_str_tag(constants.GRPC_STATUS_CODE_KEY, code)
    span._set_str_tag(ERROR_MSG, await call.details())
    span._set_str_tag(ERROR_TYPE, code)
    span.finish()


class _ClientInterceptor:
    def __init__(self, pin, host, port):
        # type: (Pin, str, int) -> None
        self._pin = pin
        self._host = host
        self._port = port

    def _intercept_client_call(self, method_kind, client_call_details):
        # type: (str, aio.ClientCallDetails) -> Tuple[Span, aio.ClientCallDetails]
        tracer = self._pin.tracer

        method_as_str = client_call_details.method.decode()
        span = tracer.trace(
            "grpc",
            span_type=SpanTypes.GRPC,
            service=trace_utils.ext_service(self._pin, config.grpc_aio_client),
            resource=method_as_str,
        )
        span.set_tag(SPAN_MEASURED_KEY)

        utils.set_grpc_method_meta(span, method_as_str, method_kind)
        utils.set_grpc_client_meta(span, self._host, self._port)
        span._set_str_tag(constants.GRPC_SPAN_KIND_KEY, constants.GRPC_SPAN_KIND_VALUE_CLIENT)

        sample_rate = config.grpc_aio_client.get_analytics_sample_rate()
        if sample_rate is not None:
            span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, sample_rate)

        # inject tags from pin
        if self._pin.tags:
            span.set_tags(self._pin.tags)

        # propagate distributed tracing headers if available
        headers = {}
        if config.grpc_aio_client.distributed_tracing_enabled:
            HTTPPropagator.inject(span.context, headers)

        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.extend(headers.items())

        client_call_details = aio.ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            metadata,
            client_call_details.credentials,
            client_call_details.wait_for_ready,
        )

        return span, client_call_details

    # NOTE: Since this function is executed as an async generator when the RPC is called,
    # `continuation` must be called before the RPC.
    async def _wrap_stream_response(
        self,
        call,  # type: Union[aio.StreamStreamCall, aio.UnaryStreamCall]
        span,  # type: Span
    ):
        # type: (...) -> ResponseIterableType
        try:
            async for response in call:
                yield response
            code = await call.code()
            details = await call.details()
            # NOTE: The callback is registered after the iteration is done,
            # otherwise `call.code()` and `call.details()` block indefinitely.
            call.add_done_callback(_done_callback(span, code, details))
        except aio.AioRpcError as rpc_error:
            # NOTE: We can also handle the error in done callbacks,
            # but reuse this error handling function used in unary response RPCs.
            _handle_rpc_error(span, rpc_error)
            raise
        except asyncio.CancelledError:
            # NOTE: We can't handle the cancelled error in done callbacks
            # because they cannot handle awaitable functions.
            await _handle_cancelled_error(call, span)
            raise

    # NOTE: `continuation` must be called inside of this function to catch exceptions.
    async def _wrap_unary_response(
        self,
        continuation,  # type: Callable[[], Union[aio.StreamUnaryCall, aio.UnaryUnaryCall]]
        span,  # type: Span
    ):
        # type: (...) -> Union[aio.StreamUnaryCall, aio.UnaryUnaryCall]
        try:
            call = await continuation()
            code = await call.code()
            details = await call.details()
            # NOTE: As both `code` and `details` are available after the RPC is done (= we get `call` object),
            # and we can't call awaitable functions inside the non-async callback,
            # there is no other way but to register the callback here.
            call.add_done_callback(_done_callback(span, code, details))
            return call
        except aio.AioRpcError as rpc_error:
            # NOTE: `AioRpcError` is raised in `await continuation(...)`
            # and `call` object is not assigned yet in that case.
            # So we can't handle the error in done callbacks.
            _handle_rpc_error(span, rpc_error)
            raise


class _UnaryUnaryClientInterceptor(aio.UnaryUnaryClientInterceptor, _ClientInterceptor):
    async def intercept_unary_unary(
        self,
        continuation,  # type: Callable[[aio.ClientCallDetails, RequestType], aio.UnaryUnaryCall]
        client_call_details,  # type: aio.ClientCallDetails
        request,  # type: RequestType
    ):
        # type: (...) -> Union[aio.UnaryUnaryCall, ResponseType]
        span, client_call_details = self._intercept_client_call(
            constants.GRPC_METHOD_KIND_UNARY,
            client_call_details,
        )
        continuation_with_args = functools.partial(continuation, client_call_details, request)
        return await self._wrap_unary_response(continuation_with_args, span)


class _UnaryStreamClientInterceptor(aio.UnaryStreamClientInterceptor, _ClientInterceptor):
    async def intercept_unary_stream(
        self,
        continuation,  # type: Callable[[aio.ClientCallDetails, RequestType], aio.UnaryStreamCall]
        client_call_details,  # type: aio.ClientCallDetails
        request,  # type: RequestType
    ):
        # type: (...) -> Union[aio.UnaryStreamCall, ResponseIterableType]
        span, client_call_details = self._intercept_client_call(
            constants.GRPC_METHOD_KIND_SERVER_STREAMING,
            client_call_details,
        )
        call = await continuation(client_call_details, request)
        return self._wrap_stream_response(call, span)


class _StreamUnaryClientInterceptor(aio.StreamUnaryClientInterceptor, _ClientInterceptor):
    async def intercept_stream_unary(
        self,
        continuation,  # type: Callable[[aio.ClientCallDetails, RequestType], aio.StreamUnaryCall]
        client_call_details,  # type: aio.ClientCallDetails
        request_iterator,  # type: RequestIterableType
    ):
        # type: (...) -> aio.StreamUnaryCall
        span, client_call_details = self._intercept_client_call(
            constants.GRPC_METHOD_KIND_CLIENT_STREAMING,
            client_call_details,
        )
        continuation_with_args = functools.partial(continuation, client_call_details, request_iterator)
        return await self._wrap_unary_response(continuation_with_args, span)


class _StreamStreamClientInterceptor(aio.StreamStreamClientInterceptor, _ClientInterceptor):
    async def intercept_stream_stream(
        self,
        continuation,  # type: Callable[[aio.ClientCallDetails, RequestType], aio.StreamStreamCall]
        client_call_details,  # type: aio.ClientCallDetails
        request_iterator,  # type: RequestIterableType
    ):
        # type: (...) -> Union[aio.StreamStreamCall, ResponseIterableType]
        span, client_call_details = self._intercept_client_call(
            constants.GRPC_METHOD_KIND_BIDI_STREAMING,
            client_call_details,
        )
        call = await continuation(client_call_details, request_iterator)
        return self._wrap_stream_response(call, span)



import asyncio


if hasattr(asyncio, "current_task"):

    def asyncio_current_task():
        try:
            return asyncio.current_task()
        except RuntimeError:
            return None


else:

    def asyncio_current_task():
        return asyncio.Task.current_task()



from contextlib import contextmanager

import mock
import pytest

from ddtrace.internal.metrics import Metrics


@contextmanager
def mock_metrics(namespace=None):
    metrics = Metrics(namespace=namespace)
    try:
        metrics._client = mock.Mock()
        metrics.enable()
        yield metrics
    finally:
        metrics.disable()


@pytest.mark.parametrize("namespace", [None, "namespace"])
def test_metrics_names(namespace):
    PI = 3.14159265359
    with mock_metrics(namespace) as metrics:
        client = metrics._client
        m = metrics.get_meter("foo.bar")
        m.increment("my.counter")
        m.distribution("my.dist", PI)

    # The meter gets disabled when the context manager returns, so the client
    # should not have these calls
    m.increment("my.counter2")
    m.distribution("my.dist2", PI)

    assert client.increment.mock_calls == [mock.call("foo.bar.my.counter", 1.0, None)]
    assert client.distribution.mock_calls == [mock.call("foo.bar.my.dist", PI, None)]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tests/contrib/grpc/hello.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1etests/contrib/grpc/hello.proto\x12\nhelloworld"\x1c\n\x0cHelloRequest\x12\x0c\n\x04name\x18\x01 \x01(\t"\x1d\n\nHelloReply\x12\x0f\n\x07message\x18\x01 \x01(\t2\xa2\x02\n\x05Hello\x12>\n\x08SayHello\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply"\x00\x12\x45\n\rSayHelloTwice\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply"\x00\x30\x01\x12L\n\x12SayHelloRepeatedly\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply"\x00(\x01\x30\x01\x12\x44\n\x0cSayHelloLast\x12\x18.helloworld.HelloRequest\x1a\x16.helloworld.HelloReply"\x00(\x01\x62\x06proto3'
)


_HELLOREQUEST = DESCRIPTOR.message_types_by_name["HelloRequest"]
_HELLOREPLY = DESCRIPTOR.message_types_by_name["HelloReply"]
HelloRequest = _reflection.GeneratedProtocolMessageType(
    "HelloRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _HELLOREQUEST,
        "__module__": "tests.contrib.grpc.hello_pb2"
        # @@protoc_insertion_point(class_scope:helloworld.HelloRequest)
    },
)
_sym_db.RegisterMessage(HelloRequest)

HelloReply = _reflection.GeneratedProtocolMessageType(
    "HelloReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _HELLOREPLY,
        "__module__": "tests.contrib.grpc.hello_pb2"
        # @@protoc_insertion_point(class_scope:helloworld.HelloReply)
    },
)
_sym_db.RegisterMessage(HelloReply)

_HELLO = DESCRIPTOR.services_by_name["Hello"]
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _HELLOREQUEST._serialized_start = 46
    _HELLOREQUEST._serialized_end = 74
    _HELLOREPLY._serialized_start = 76
    _HELLOREPLY._serialized_end = 105
    _HELLO._serialized_start = 108
    _HELLO._serialized_end = 398
# @@protoc_insertion_point(module_scope)



import os
import sys
import typing


def get_application_name():
    # type: () -> typing.Optional[str]
    """Attempts to find the application name using system arguments."""
    try:
        import __main__

        name = __main__.__file__
    except (ImportError, AttributeError):
        try:
            name = sys.argv[0]
        except (AttributeError, IndexError):
            return None

    return os.path.basename(name)



import subprocess
import sys

import pytest
import requests
from requests import Session
from requests.exceptions import InvalidURL
from requests.exceptions import MissingSchema
import six

from ddtrace import Pin
from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.constants import ERROR_MSG
from ddtrace.constants import ERROR_STACK
from ddtrace.constants import ERROR_TYPE
from ddtrace.contrib.requests import patch
from ddtrace.contrib.requests import unpatch
from ddtrace.contrib.requests.connection import _extract_hostname
from ddtrace.contrib.requests.connection import _extract_query_string
from ddtrace.ext import http
from tests.opentracer.utils import init_tracer
from tests.utils import TracerTestCase
from tests.utils import assert_is_measured
from tests.utils import assert_span_http_status_code
from tests.utils import override_global_tracer


# socket name comes from https://english.stackexchange.com/a/44048
SOCKET = "httpbin.org"
URL_200 = "http://{}/status/200".format(SOCKET)
URL_500 = "http://{}/status/500".format(SOCKET)


class BaseRequestTestCase(object):
    """Create a traced Session, patching during the setUp and
    unpatching after the tearDown
    """

    def setUp(self):
        super(BaseRequestTestCase, self).setUp()

        patch()
        self.session = Session()
        setattr(self.session, "datadog_tracer", self.tracer)

    def tearDown(self):
        unpatch()

        super(BaseRequestTestCase, self).tearDown()


class TestRequests(BaseRequestTestCase, TracerTestCase):
    def test_resource_path(self):
        out = self.session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag("http.url") == URL_200

    def test_tracer_disabled(self):
        # ensure all valid combinations of args / kwargs work
        self.tracer.enabled = False
        out = self.session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert len(spans) == 0

    def test_args_kwargs(self):
        # ensure all valid combinations of args / kwargs work
        url = URL_200
        method = "GET"
        inputs = [
            ([], {"method": method, "url": url}),
            ([method], {"url": url}),
            ([method, url], {}),
        ]

        for args, kwargs in inputs:
            # ensure a traced request works with these args
            out = self.session.request(*args, **kwargs)
            assert out.status_code == 200
            # validation
            spans = self.pop_spans()
            assert len(spans) == 1
            s = spans[0]
            assert s.get_tag(http.METHOD) == "GET"
            assert_span_http_status_code(s, 200)

    def test_untraced_request(self):
        # ensure the unpatch removes tracing
        unpatch()
        untraced = Session()

        out = untraced.get(URL_200)
        assert out.status_code == 200
        # validation
        spans = self.pop_spans()
        assert len(spans) == 0

    def test_double_patch(self):
        # ensure that double patch doesn't duplicate instrumentation
        patch()
        session = Session()
        setattr(session, "datadog_tracer", self.tracer)

        out = session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert len(spans) == 1

    def test_200(self):
        out = self.session.get(URL_200)
        assert out.status_code == 200
        # validation
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert_is_measured(s)
        assert s.get_tag(http.METHOD) == "GET"
        assert_span_http_status_code(s, 200)
        assert s.error == 0
        assert s.span_type == "http"
        assert http.QUERY_STRING not in s.get_tags()

    def test_200_send(self):
        # when calling send directly
        req = requests.Request(url=URL_200, method="GET")
        req = self.session.prepare_request(req)

        out = self.session.send(req)
        assert out.status_code == 200
        # validation
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert_is_measured(s)
        assert s.get_tag(http.METHOD) == "GET"
        assert_span_http_status_code(s, 200)
        assert s.error == 0
        assert s.span_type == "http"

    def test_200_query_string(self):
        # ensure query string is removed before adding url to metadata
        query_string = "key=value&key2=value2"
        with self.override_http_config("requests", dict(trace_query_string=True)):
            out = self.session.get(URL_200 + "?" + query_string)
        assert out.status_code == 200
        # validation
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert_is_measured(s)
        assert s.get_tag(http.METHOD) == "GET"
        assert_span_http_status_code(s, 200)
        assert s.get_tag(http.URL) == URL_200 + "?" + query_string
        assert s.error == 0
        assert s.span_type == "http"
        assert s.get_tag(http.QUERY_STRING) == query_string

    def test_requests_module_200(self):
        # ensure the requests API is instrumented even without
        # using a `Session` directly
        with override_global_tracer(self.tracer):
            out = requests.get(URL_200)
            assert out.status_code == 200
            # validation
            spans = self.pop_spans()
            assert len(spans) == 1
            s = spans[0]

            assert_is_measured(s)
            assert s.get_tag(http.METHOD) == "GET"
            assert_span_http_status_code(s, 200)
            assert s.error == 0
            assert s.span_type == "http"

    def test_post_500(self):
        out = self.session.post(URL_500)
        # validation
        assert out.status_code == 500
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert_is_measured(s)
        assert s.get_tag(http.METHOD) == "POST"
        assert_span_http_status_code(s, 500)
        assert s.error == 1

    def test_non_existant_url(self):
        try:
            self.session.get("http://doesnotexist.google.com")
        except Exception:
            pass
        else:
            assert 0, "expected error"

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert_is_measured(s)
        assert s.get_tag(http.METHOD) == "GET"
        assert s.error == 1
        assert "Failed to establish a new connection" in s.get_tag(ERROR_MSG)
        assert "Failed to establish a new connection" in s.get_tag(ERROR_STACK)
        assert "Traceback (most recent call last)" in s.get_tag(ERROR_STACK)
        assert "requests.exception" in s.get_tag(ERROR_TYPE)

    def test_500(self):
        out = self.session.get(URL_500)
        assert out.status_code == 500

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert_is_measured(s)
        assert s.get_tag(http.METHOD) == "GET"
        assert_span_http_status_code(s, 500)
        assert s.error == 1

    def test_default_service_name(self):
        # ensure a default service name is set
        out = self.session.get(URL_200)
        assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert s.service == "requests"

    def test_user_set_service_name(self):
        # ensure a service name set by the user has precedence
        cfg = config.get_from(self.session)
        cfg["service_name"] = "clients"
        out = self.session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.service == "clients"

    def test_user_set_service(self):
        # ensure a service name set by the user has precedence
        cfg = config.get_from(self.session)
        cfg["service"] = "clients"
        out = self.session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.service == "clients"

    def test_parent_service_name_precedence(self):
        # span should not inherit the parent's service name
        # as all outbound requests should go to a difference service
        with self.tracer.trace("parent.span", service="web"):
            out = self.session.get(URL_200)
            assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 2
        s = spans[1]

        assert s.name == "requests.request"
        assert s.service == "requests"

    def test_parent_without_service_name(self):
        # ensure the default value is used if the parent
        # doesn't have a service
        with self.tracer.trace("parent.span"):
            out = self.session.get(URL_200)
            assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 2
        s = spans[1]

        assert s.name == "requests.request"
        assert s.service == "requests"

    def test_user_service_name_precedence(self):
        # ensure the user service name takes precedence over
        # the parent Span
        cfg = config.get_from(self.session)
        cfg["service_name"] = "clients"
        with self.tracer.trace("parent.span", service="web"):
            out = self.session.get(URL_200)
            assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 2
        s = spans[1]

        assert s.name == "requests.request"
        assert s.service == "clients"

    def test_split_by_domain_with_ampersat(self):
        # Regression test for: https://github.com/DataDog/dd-trace-py/issues/4062
        # ensure a service name is generated by the domain name
        cfg = config.get_from(self.session)
        cfg["split_by_domain"] = True
        # domain name should take precedence over monkey_service
        cfg["service_name"] = "monkey_service"
        url = URL_200 + "?email=monkey_monkey@zoo_mail.ca"

        out = self.session.get(url)
        assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert s.service == "httpbin.org"

    def test_split_by_domain(self):
        # ensure a service name is generated by the domain name
        # of the ongoing call
        cfg = config.get_from(self.session)
        cfg["split_by_domain"] = True
        out = self.session.get(URL_200)
        assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert s.service == "httpbin.org"

    def test_split_by_domain_precedence(self):
        # ensure the split by domain has precedence all the time
        cfg = config.get_from(self.session)
        cfg["split_by_domain"] = True
        cfg["service_name"] = "intake"
        out = self.session.get(URL_200)
        assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert s.service == "httpbin.org"

    def test_split_by_domain_wrong(self):
        # ensure the split by domain doesn't crash in case of a wrong URL;
        # in that case, no spans are created
        cfg = config.get_from(self.session)
        cfg["split_by_domain"] = True
        with pytest.raises((MissingSchema, InvalidURL)):
            self.session.get("http:/some>thing")

        # We are wrapping `requests.Session.send` and this error gets thrown before that function
        spans = self.pop_spans()
        assert len(spans) == 0

    def test_split_by_domain_remove_auth_in_url(self):
        # ensure that auth details are stripped from URL
        cfg = config.get_from(self.session)
        cfg["split_by_domain"] = True
        out = self.session.get("http://user:pass@httpbin.org")
        assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert s.service == "httpbin.org"

    def test_split_by_domain_includes_port(self):
        # ensure that port is included if present in URL
        cfg = config.get_from(self.session)
        cfg["split_by_domain"] = True
        out = self.session.get("http://httpbin.org:80")
        assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert s.service == "httpbin.org:80"

    def test_split_by_domain_includes_port_path(self):
        # ensure that port is included if present in URL but not path
        cfg = config.get_from(self.session)
        cfg["split_by_domain"] = True
        out = self.session.get("http://httpbin.org:80/anything/v1/foo")
        assert out.status_code == 200

        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]

        assert s.service == "httpbin.org:80"

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_REQUESTS_SERVICE="override"))
    def test_global_config_service_env_precedence(self):
        out = self.session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert spans[0].service == "override"

        cfg = config.get_from(self.session)
        cfg["service"] = "override2"
        out = self.session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert spans[0].service == "override2"

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_REQUESTS_SERVICE="override"))
    def test_global_config_service_env(self):
        out = self.session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert spans[0].service == "override"

    def test_global_config_service(self):
        with self.override_config("requests", dict(service="override")):
            out = self.session.get(URL_200)
        assert out.status_code == 200
        spans = self.pop_spans()
        assert spans[0].service == "override"

    def test_200_ot(self):
        """OpenTracing version of test_200."""

        ot_tracer = init_tracer("requests_svc", self.tracer)

        with ot_tracer.start_active_span("requests_get"):
            out = self.session.get(URL_200)
            assert out.status_code == 200

        # validation
        spans = self.pop_spans()
        assert len(spans) == 2

        ot_span, dd_span = spans

        # confirm the parenting
        assert ot_span.parent_id is None
        assert dd_span.parent_id == ot_span.span_id

        assert ot_span.name == "requests_get"
        assert ot_span.service == "requests_svc"

        assert_is_measured(dd_span)
        assert dd_span.get_tag(http.METHOD) == "GET"
        assert_span_http_status_code(dd_span, 200)
        assert dd_span.error == 0
        assert dd_span.span_type == "http"

    def test_request_and_response_headers(self):
        # Disabled when not configured
        self.session.get(URL_200, headers={"my-header": "my_value"})
        spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag("http.request.headers.my-header") is None
        assert s.get_tag("http.response.headers.access-control-allow-origin") is None

        # Enabled when explicitly configured
        with self.override_config("requests", {}):
            config.requests.http.trace_headers(["my-header", "access-control-allow-origin"])
            self.session.get(URL_200, headers={"my-header": "my_value"})
            spans = self.pop_spans()
        assert len(spans) == 1
        s = spans[0]
        assert s.get_tag("http.request.headers.my-header") == "my_value"
        assert s.get_tag("http.response.headers.access-control-allow-origin") == "*"

    def test_analytics_integration_default(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set
                We expect the root span to have the appropriate tag
        """
        self.session.get(URL_200)

        spans = self.pop_spans()
        self.assertEqual(len(spans), 1)
        s = spans[0]
        self.assertIsNone(s.get_metric(ANALYTICS_SAMPLE_RATE_KEY))

    def test_analytics_integration_disabled(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set
                We expect the root span to have the appropriate tag
        """
        with self.override_config("requests", dict(analytics_enabled=False, analytics_sample_rate=0.5)):
            self.session.get(URL_200)

        spans = self.pop_spans()
        self.assertEqual(len(spans), 1)
        s = spans[0]
        self.assertIsNone(s.get_metric(ANALYTICS_SAMPLE_RATE_KEY))

    def test_analytics_integration_on(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set
                We expect the root span to have the appropriate tag
        """
        with self.override_config("requests", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
            self.session.get(URL_200)

        spans = self.pop_spans()
        self.assertEqual(len(spans), 1)
        s = spans[0]
        self.assertEqual(s.get_metric(ANALYTICS_SAMPLE_RATE_KEY), 0.5)

    def test_analytics_integration_on_using_pin(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set
                We expect the root span to have the appropriate tag
        """
        pin = Pin(
            service=__name__,
            _config={
                "service_name": __name__,
                "distributed_tracing": False,
                "split_by_domain": False,
                "analytics_enabled": True,
                "analytics_sample_rate": 0.5,
            },
        )
        pin.onto(self.session)
        self.session.get(URL_200)

        spans = self.pop_spans()
        self.assertEqual(len(spans), 1)
        s = spans[0]
        self.assertEqual(s.get_metric(ANALYTICS_SAMPLE_RATE_KEY), 0.5)

    def test_analytics_integration_on_using_pin_default(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set
                We expect the root span to have the appropriate tag
        """
        pin = Pin(
            service=__name__,
            _config={
                "service_name": __name__,
                "distributed_tracing": False,
                "split_by_domain": False,
                "analytics_enabled": True,
            },
        )
        pin.onto(self.session)
        self.session.get(URL_200)

        spans = self.pop_spans()
        self.assertEqual(len(spans), 1)
        s = spans[0]
        self.assertEqual(s.get_metric(ANALYTICS_SAMPLE_RATE_KEY), 1.0)


def test_traced_session_no_patch_all(tmpdir):
    f = tmpdir.join("test.py")
    f.write(
        """
import mock
import ddtrace
from ddtrace.contrib.requests import TracedSession

# disable tracer writing to agent
# FIXME: Remove use of this internal attribute of Tracer to disable writer
ddtrace.tracer._writer.flush_queue = mock.Mock(return_value=None)

session = TracedSession()
session.get("http://httpbin.org/status/200")
""".lstrip()
    )
    p = subprocess.Popen(
        [sys.executable, "test.py"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=str(tmpdir),
    )
    p.wait()
    assert p.stderr.read() == six.b("")
    assert p.stdout.read() == six.b("")
    assert p.returncode == 0


@pytest.mark.parametrize(
    "uri,hostname",
    [
        ("http://localhost:8080", "localhost:8080"),
        ("http://localhost:8080/", "localhost:8080"),
        ("http://localhost", "localhost"),
        ("http://localhost/", "localhost"),
        ("http://asd:wwefwf@localhost:8080", "localhost:8080"),
        ("http://asd:wwefwf@localhost:8080/", "localhost:8080"),
        ("http://asd:wwefwf@localhost:8080/path", "localhost:8080"),
        ("http://asd:wwefwf@localhost:8080/path?query", "localhost:8080"),
        ("http://asd:wwefwf@localhost:8080/path?query#fragment", "localhost:8080"),
        ("http://asd:wwefwf@localhost:8080/path#frag?ment", "localhost:8080"),
        ("http://localhost:8080/path#frag?ment", "localhost:8080"),
        ("http://localhost/path#frag?ment", "localhost"),
    ],
)
def test_extract_hostname(uri, hostname):
    assert _extract_hostname(uri) == hostname


def test_extract_hostname_invalid_port():
    if sys.version_info < (3, 6):
        assert _extract_hostname("http://localhost:-1/") == "localhost"
    else:
        assert _extract_hostname("http://localhost:-1/") == "localhost:?"


@pytest.mark.parametrize(
    "uri,qs",
    [
        ("http://localhost:8080", None),
        ("http://localhost", None),
        ("http://asd:wwefwf@localhost:8080", None),
        ("http://asd:wwefwf@localhost:8080/path", None),
        ("http://asd:wwefwf@localhost:8080/path?query", "query"),
        ("http://asd:wwefwf@localhost:8080/path?query#fragment", "query"),
        ("http://asd:wwefwf@localhost:8080/path#frag?ment", None),
        ("http://localhost:8080/path#frag?ment", None),
        ("http://localhost/path#frag?ment", None),
        ("http://localhost?query", "query"),
    ],
)
def test_extract_query_string(uri, qs):
    assert _extract_query_string(uri) == qs



import sys
import warnings


def gevent_patch_all(event):
    if "ddtrace" in sys.modules:
        warnings.warn(
            "Loading ddtrace before using gevent monkey patching is not supported "
            "and is likely to break the application. "
            "Use `DD_GEVENT_PATCH_ALL=true ddtrace-run` to fix this or "
            "import `ddtrace` after `gevent.monkey.patch_all()` has been called.",
            RuntimeWarning,
        )



import re

from ...ext import elasticsearch as metadata


# Replace any ID
ID_REGEXP = re.compile(r"/([0-9]+)([/\?]|$)")
ID_PLACEHOLDER = r"/?\2"

# Remove digits from potential timestamped indexes (should be an option).
# For now, let's say 2+ digits
INDEX_REGEXP = re.compile(r"[0-9]{2,}")
INDEX_PLACEHOLDER = r"?"


def quantize(span):
    """Quantize an elasticsearch span

    We want to extract a meaningful `resource` from the request.
    We do it based on the method + url, with some cleanup applied to the URL.

    The URL might a ID, but also it is common to have timestamped indexes.
    While the first is easy to catch, the second should probably be configurable.

    All of this should probably be done in the Agent. Later.
    """
    url = span.get_tag(metadata.URL)
    method = span.get_tag(metadata.METHOD)

    quantized_url = ID_REGEXP.sub(ID_PLACEHOLDER, url)
    quantized_url = INDEX_REGEXP.sub(INDEX_PLACEHOLDER, quantized_url)

    span.resource = "{method} {url}".format(method=method, url=quantized_url)

    return span



from typing import TYPE_CHECKING

from ddtrace.provider import _DD_CONTEXTVAR
from ddtrace.span import Span


if TYPE_CHECKING:
    from typing import Any
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Union

    from ddtrace.context import Context


def get_item(key, span=None):
    # type: (str, Optional[Span]) -> Optional[Any]
    """Get and item from the context of a trace."""
    ctx = span if span is not None else _DD_CONTEXTVAR.get()
    if not isinstance(ctx, Span) or ctx._local_root is None:
        raise ValueError("No context found")
    return ctx._local_root._get_ctx_item(key)


def get_items(keys, span=None):
    # type: (List[str], Optional[Span]) -> List[Optional[Any]]
    """Get multiple items from the context of a trace."""
    ctx = span if span is not None else _DD_CONTEXTVAR.get()  # type: Optional[Union[Context, Span]]
    if not isinstance(ctx, Span) or ctx._local_root is None:
        raise ValueError("No context found")
    return [ctx._local_root._get_ctx_item(k) for k in keys]


def set_item(key, val, span=None):
    # type: (str, Any, Optional[Span]) -> None
    """Set an item in the context of a trace."""
    ctx = span if span is not None else _DD_CONTEXTVAR.get()  # type: Optional[Union[Context, Span]]
    if not isinstance(ctx, Span) or ctx._local_root is None:
        raise ValueError("No context found")
    ctx._local_root._set_ctx_item(key, val)


def set_items(kvs, span=None):
    # type: (Dict[str, Any], Optional[Span]) -> None
    """Set multiple items in the context of a trace."""
    ctx = span if span is not None else _DD_CONTEXTVAR.get()  # type: Optional[Union[Context, Span]]
    if not isinstance(ctx, Span) or ctx._local_root is None:
        raise ValueError("No context found")
    ctx._local_root._set_ctx_items(kvs)



import sys
from types import CoroutineType
from typing import List

from ddtrace.debugging._snapshot.collector import SnapshotContext
from ddtrace.internal import compat


async def dd_coroutine_wrapper(coro, contexts):
    # type: (CoroutineType, List[SnapshotContext]) -> CoroutineType
    start_time = compat.monotonic_ns()
    try:
        retval = await coro
        end_time = compat.monotonic_ns()
        exc_info = (None, None, None)
    except Exception:
        end_time = compat.monotonic_ns()
        retval = None
        exc_info = sys.exc_info()  # type: ignore[assignment]

    for context in contexts:
        context.exit(retval, exc_info, end_time - start_time)

    _, exc, _ = exc_info
    if exc is not None:
        raise exc

    return retval



GC_COUNT_GEN0 = "runtime.python.gc.count.gen0"
GC_COUNT_GEN1 = "runtime.python.gc.count.gen1"
GC_COUNT_GEN2 = "runtime.python.gc.count.gen2"

THREAD_COUNT = "runtime.python.thread_count"
MEM_RSS = "runtime.python.mem.rss"
# `runtime.python.cpu.time.sys` metric is used to auto-enable runtime metrics dashboards in DD backend
CPU_TIME_SYS = "runtime.python.cpu.time.sys"
CPU_TIME_USER = "runtime.python.cpu.time.user"
CPU_PERCENT = "runtime.python.cpu.percent"
CTX_SWITCH_VOLUNTARY = "runtime.python.cpu.ctx_switch.voluntary"
CTX_SWITCH_INVOLUNTARY = "runtime.python.cpu.ctx_switch.involuntary"

GC_RUNTIME_METRICS = set([GC_COUNT_GEN0, GC_COUNT_GEN1, GC_COUNT_GEN2])

PSUTIL_RUNTIME_METRICS = set(
    [THREAD_COUNT, MEM_RSS, CTX_SWITCH_VOLUNTARY, CTX_SWITCH_INVOLUNTARY, CPU_TIME_SYS, CPU_TIME_USER, CPU_PERCENT]
)

DEFAULT_RUNTIME_METRICS = GC_RUNTIME_METRICS | PSUTIL_RUNTIME_METRICS

SERVICE = "service"
ENV = "env"
LANG_INTERPRETER = "lang_interpreter"
LANG_VERSION = "lang_version"
LANG = "lang"
TRACER_VERSION = "tracer_version"

TRACER_TAGS = set([SERVICE, ENV])

PLATFORM_TAGS = set([LANG_INTERPRETER, LANG_VERSION, LANG, TRACER_VERSION])

DEFAULT_RUNTIME_TAGS = TRACER_TAGS | PLATFORM_TAGS



"""
The aredis integration traces aredis requests.


Enabling
~~~~~~~~

The aredis integration is enabled automatically when using
:ref:`ddtrace-run<ddtracerun>` or :func:`patch_all()<ddtrace.patch_all>`.

Or use :func:`patch()<ddtrace.patch>` to manually enable the integration::

    from ddtrace import patch
    patch(aredis=True)


Global Configuration
~~~~~~~~~~~~~~~~~~~~

.. py:data:: ddtrace.config.aredis["service"]

   The service name reported by default for aredis traces.

   This option can also be set with the ``DD_AREDIS_SERVICE`` environment
   variable.

   Default: ``"redis"``


Instance Configuration
~~~~~~~~~~~~~~~~~~~~~~

To configure particular aredis instances use the :class:`Pin <ddtrace.Pin>` API::

    import aredis
    from ddtrace import Pin

    client = aredis.StrictRedis(host="localhost", port=6379)

    # Override service name for this instance
    Pin.override(client, service="my-custom-queue")

    # Traces reported for this client will now have "my-custom-queue"
    # as the service name.
    async def example():
        await client.get("my-key")
"""

from ...internal.utils.importlib import require_modules


required_modules = ["aredis", "aredis.client"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch

        __all__ = ["patch"]



"""
Class based views used for Django tests.
"""

from functools import partial

from django.contrib.auth.models import User
from django.contrib.syndication.views import Feed
from django.http import Http404
from django.http import HttpResponse
from django.template import loader
from django.template.response import TemplateResponse
from django.utils.safestring import mark_safe
from django.views.generic import ListView
from django.views.generic import TemplateView
from django.views.generic import View

from ddtrace import tracer
from ddtrace.contrib.trace_utils import set_user


class UserList(ListView):
    model = User
    template_name = "users_list.html"


class TemplateCachedUserList(ListView):
    model = User
    template_name = "cached_list.html"


class SafeTemplateUserList(ListView):
    model = User
    template_name = mark_safe("cached_list.html")


class BasicView(View):
    def get(self, request):
        return HttpResponse("")

    def post(self, request):
        return HttpResponse("")

    def delete(self, request):
        return HttpResponse("")

    def head(self, request):
        return HttpResponse("")


class ForbiddenView(TemplateView):
    def get(self, request, *args, **kwargs):
        return HttpResponse(status=403)


class StaticMethodView(View):
    @staticmethod
    def get(request):
        return HttpResponse("")


def function_view(request):
    return HttpResponse(status=200)


def error_500(request):
    raise Exception("Error 500")


class FeedView(Feed):
    """
    A callable view that is part of the Django framework
    """

    title = "Police beat site news"
    link = "/sitenews/"
    description = "Updates on changes and additions to police beat central."

    def items(self):
        return []

    def item_title(self, item):
        return "empty"

    def item_description(self, item):
        return "empty"


partial_view = partial(function_view)

# disabling flake8 test below, yes, declaring a func like this is bad, we know
lambda_view = lambda request: function_view(request)  # NOQA


def index(request):
    response = HttpResponse("Hello, test app.")
    response["my-response-header"] = "my_response_value"
    return response


def alter_resource(request):
    root = tracer.current_root_span()
    root.resource = "custom django.request resource"

    return HttpResponse("")


def template_view(request):
    """
    View that uses a template instance
    """
    template = loader.select_template(["basic.html"])
    return TemplateResponse(request, template)


def template_simple_view(request):
    """
    Basic django templated view
    """
    return TemplateResponse(request, "basic.html")


def template_list_view(request):
    """
    For testing resolving a list of templates
    """
    return TemplateResponse(request, ["doesntexist.html", "basic.html"])


class CustomDispatchMixin(View):
    def dispatch(self, request):
        self.dispatch_call_counter += 1
        return super(CustomDispatchMixin, self).dispatch(request)


class AnotherCustomDispatchMixin(View):
    def dispatch(self, request):
        self.dispatch_call_counter += 1
        return super(AnotherCustomDispatchMixin, self).dispatch(request)


class ComposedTemplateView(TemplateView, CustomDispatchMixin, AnotherCustomDispatchMixin):
    template_name = "custom_dispatch.html"
    dispatch_call_counter = 0

    def get_context_data(self, **kwargs):
        context = super(ComposedTemplateView, self).get_context_data(**kwargs)
        context["dispatch_call_counter"] = self.dispatch_call_counter
        return context


class CustomGetView(View):
    def get(self, request):
        return HttpResponse("custom get")


class ComposedGetView(CustomGetView, CustomDispatchMixin):
    dispatch_call_counter = 0

    def get(self, request):
        if self.dispatch_call_counter == 1:
            return super(ComposedGetView, self).get(request)
        raise Exception("Custom dispatch not called.")


DISPATCH_CALLED = False


class CustomDispatchView(View):
    def dispatch(self, request):
        global DISPATCH_CALLED
        DISPATCH_CALLED = True
        return super(CustomDispatchView, self).dispatch(request)


class ComposedView(TemplateView, CustomDispatchView):
    template_name = "custom_dispatch.html"


def not_found_view(request):
    raise Http404("DNE")


def path_params_view(request, year, month):
    return HttpResponse(status=200)


def identify(request):
    set_user(
        tracer,
        user_id="usr.id",
        email="usr.email",
        name="usr.name",
        session_id="usr.session_id",
        role="usr.role",
        scope="usr.scope",
    )
    return HttpResponse(status=200)


def body_view(request):
    # Django >= 3
    if hasattr(request, "headers"):
        content_type = request.headers["Content-Type"]
    else:
        # Django < 3
        content_type = request.META["CONTENT_TYPE"]
    if content_type in ("application/json", "application/xml", "text/xml"):
        data = request.body
        return HttpResponse(data, status=200)
    else:
        data = request.POST
        return HttpResponse(str(dict(data)), status=200)



import inspect
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Iterable
from typing import Union

import grpc
from grpc import aio
from grpc.aio._typing import RequestIterableType
from grpc.aio._typing import RequestType
from grpc.aio._typing import ResponseIterableType
from grpc.aio._typing import ResponseType

from ddtrace import Pin
from ddtrace import Span
from ddtrace import config
from ddtrace.vendor import wrapt

from .. import trace_utils
from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import ERROR_MSG
from ...constants import ERROR_TYPE
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...internal.compat import to_unicode
from ..grpc import constants
from ..grpc.utils import set_grpc_method_meta


Continuation = Callable[[grpc.HandlerCallDetails], Awaitable[grpc.RpcMethodHandler]]


# Used to get a status code from integer
# as `grpc._cython.cygrpc._ServicerContext.code()` returns an integer.
_INT2CODE = {s.value[0]: s for s in grpc.StatusCode}


def _is_coroutine_rpc_method_handler(handler):
    # type: (grpc.RpcMethodHandler) -> bool
    if not handler.request_streaming and not handler.response_streaming:
        return inspect.iscoroutinefunction(handler.unary_unary)
    elif not handler.request_streaming and handler.response_streaming:
        return inspect.isasyncgenfunction(handler.unary_stream)
    elif handler.request_streaming and not handler.response_streaming:
        return inspect.iscoroutinefunction(handler.stream_unary)
    else:
        return inspect.isasyncgenfunction(handler.stream_stream)


def create_aio_server_interceptor(pin):
    # type: (Pin) -> _ServerInterceptor
    async def interceptor_function(
        continuation,  # type: Continuation
        handler_call_details,  # type: grpc.HandlerCallDetails
    ):
        # type: (...) -> Union[grpc.RpcMethodHandler, _TracedAioRpcMethodHandler, _TracedRpcMethodHandler, None]
        rpc_method_handler = await continuation(handler_call_details)

        # continuation returns an RpcMethodHandler instance if the RPC is
        # considered serviced, or None otherwise
        # https://grpc.github.io/grpc/python/grpc.html#grpc.ServerInterceptor.intercept_service

        if rpc_method_handler is None:
            return None

        if _is_coroutine_rpc_method_handler(rpc_method_handler):
            return _TracedAioRpcMethodHandler(pin, handler_call_details, rpc_method_handler)
        else:
            return _TracedRpcMethodHandler(pin, handler_call_details, rpc_method_handler)

    return _ServerInterceptor(interceptor_function)


def _handle_server_exception(
    servicer_context,  # type: Union[None, grpc.ServicerContext]
    span,  # type: Span
):
    # type: (...) -> None
    span.error = 1
    if servicer_context is None:
        return
    if hasattr(servicer_context, "details"):
        span._set_str_tag(ERROR_MSG, to_unicode(servicer_context.details()))
    if hasattr(servicer_context, "code") and servicer_context.code() != 0 and servicer_context.code() in _INT2CODE:
        span._set_str_tag(ERROR_TYPE, to_unicode(_INT2CODE[servicer_context.code()]))


async def _wrap_aio_stream_response(
    behavior,  # type: Callable[[Union[RequestIterableType, RequestType], aio.ServicerContext], ResponseIterableType]
    request_or_iterator,  # type: Union[RequestIterableType, RequestType]
    servicer_context,  # type: aio.ServicerContext
    span,  # type: Span
):
    # type: (...) -> ResponseIterableType
    try:
        call = behavior(request_or_iterator, servicer_context)
        async for response in call:
            yield response
    except Exception:
        span.set_traceback()
        _handle_server_exception(servicer_context, span)
        raise
    finally:
        span.finish()


async def _wrap_aio_unary_response(
    behavior,  # type: Callable[[Union[RequestIterableType, RequestType], aio.ServicerContext], Awaitable[ResponseType]]
    request_or_iterator,  # type: Union[RequestIterableType, RequestType]
    servicer_context,  # type: aio.ServicerContext
    span,  # type: Span
):
    # type: (...) -> ResponseType
    try:
        return await behavior(request_or_iterator, servicer_context)
    except Exception:
        span.set_traceback()
        _handle_server_exception(servicer_context, span)
        raise
    finally:
        span.finish()


def _wrap_stream_response(
    behavior,  # type: Callable[[Any, grpc.ServicerContext], Iterable[Any]]
    request_or_iterator,  # type: Any
    servicer_context,  # type: grpc.ServicerContext
    span,  # type: Span
):
    # type: (...) -> Iterable[Any]
    try:
        for response in behavior(request_or_iterator, servicer_context):
            yield response
    except Exception:
        span.set_traceback()
        _handle_server_exception(servicer_context, span)
        raise
    finally:
        span.finish()


def _wrap_unary_response(
    behavior,  # type: Callable[[Any, grpc.ServicerContext], Any]
    request_or_iterator,  # type: Any
    servicer_context,  # type: grpc.ServicerContext
    span,  # type: Span
):
    # type: (...) -> Any
    try:
        return behavior(request_or_iterator, servicer_context)
    except Exception:
        span.set_traceback()
        _handle_server_exception(servicer_context, span)
        raise
    finally:
        span.finish()


def _create_span(pin, handler_call_details, method_kind):
    # type: (Pin, grpc.HandlerCallDetails, str) -> Span
    tracer = pin.tracer
    headers = dict(handler_call_details.invocation_metadata)

    trace_utils.activate_distributed_headers(tracer, int_config=config.grpc_aio_server, request_headers=headers)

    span = tracer.trace(
        "grpc",
        span_type=SpanTypes.GRPC,
        service=trace_utils.int_service(pin, config.grpc_aio_server),
        resource=handler_call_details.method,
    )
    span.set_tag(SPAN_MEASURED_KEY)

    set_grpc_method_meta(span, handler_call_details.method, method_kind)
    span._set_str_tag(constants.GRPC_SPAN_KIND_KEY, constants.GRPC_SPAN_KIND_VALUE_SERVER)

    sample_rate = config.grpc_aio_server.get_analytics_sample_rate()
    if sample_rate is not None:
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, sample_rate)

    if pin.tags:
        span.set_tags(pin.tags)

    return span


class _TracedAioRpcMethodHandler(wrapt.ObjectProxy):
    def __init__(self, pin, handler_call_details, wrapped):
        # type: (Pin, grpc.HandlerCallDetails, grpc.RpcMethodHandler) -> None
        super(_TracedAioRpcMethodHandler, self).__init__(wrapped)
        self._pin = pin
        self._handler_call_details = handler_call_details

    async def unary_unary(self, request, context):
        # type: (RequestType, aio.ServicerContext) -> ResponseType
        span = _create_span(self._pin, self._handler_call_details, constants.GRPC_METHOD_KIND_UNARY)
        return await _wrap_aio_unary_response(self.__wrapped__.unary_unary, request, context, span)

    async def unary_stream(self, request, context):
        # type: (RequestType, aio.ServicerContext) -> ResponseIterableType
        span = _create_span(self._pin, self._handler_call_details, constants.GRPC_METHOD_KIND_SERVER_STREAMING)
        async for response in _wrap_aio_stream_response(self.__wrapped__.unary_stream, request, context, span):
            yield response

    async def stream_unary(self, request_iterator, context):
        # type: (RequestIterableType, aio.ServicerContext) -> ResponseType
        span = _create_span(self._pin, self._handler_call_details, constants.GRPC_METHOD_KIND_CLIENT_STREAMING)
        return await _wrap_aio_unary_response(self.__wrapped__.stream_unary, request_iterator, context, span)

    async def stream_stream(self, request_iterator, context):
        # type: (RequestIterableType, aio.ServicerContext) -> ResponseIterableType
        span = _create_span(self._pin, self._handler_call_details, constants.GRPC_METHOD_KIND_BIDI_STREAMING)
        async for response in _wrap_aio_stream_response(
            self.__wrapped__.stream_stream, request_iterator, context, span
        ):
            yield response


class _TracedRpcMethodHandler(wrapt.ObjectProxy):
    def __init__(self, pin, handler_call_details, wrapped):
        # type: (Pin, grpc.HandlerCallDetails, grpc.RpcMethodHandler) -> None
        super(_TracedRpcMethodHandler, self).__init__(wrapped)
        self._pin = pin
        self._handler_call_details = handler_call_details

    def unary_unary(self, request, context):
        # type: (Any, grpc.ServicerContext) -> Any
        span = _create_span(self._pin, self._handler_call_details, constants.GRPC_METHOD_KIND_UNARY)
        return _wrap_unary_response(self.__wrapped__.unary_unary, request, context, span)

    def unary_stream(self, request, context):
        # type: (Any, grpc.ServicerContext) -> Iterable[Any]
        span = _create_span(self._pin, self._handler_call_details, constants.GRPC_METHOD_KIND_SERVER_STREAMING)
        for response in _wrap_stream_response(self.__wrapped__.unary_stream, request, context, span):
            yield response

    def stream_unary(self, request_iterator, context):
        # type: (Iterable[Any], grpc.ServicerContext) -> Any
        span = _create_span(self._pin, self._handler_call_details, constants.GRPC_METHOD_KIND_CLIENT_STREAMING)
        return _wrap_unary_response(self.__wrapped__.stream_unary, request_iterator, context, span)

    def stream_stream(self, request_iterator, context):
        # type: (Iterable[Any], grpc.ServicerContext) -> Iterable[Any]
        span = _create_span(self._pin, self._handler_call_details, constants.GRPC_METHOD_KIND_BIDI_STREAMING)
        for response in _wrap_stream_response(self.__wrapped__.stream_stream, request_iterator, context, span):
            yield response


class _ServerInterceptor(aio.ServerInterceptor):
    def __init__(self, interceptor_function):
        self._fn = interceptor_function

    async def intercept_service(
        self,
        continuation,  # type: Continuation
        handler_call_details,  # type: grpc.HandlerCallDetails
    ):
        # type: (...) -> Union[grpc.RpcMethodHandler, _TracedAioRpcMethodHandler]
        return await self._fn(continuation, handler_call_details)



import asyncio
import sys

import aioredis

from ddtrace import config
from ddtrace.internal.utils.wrappers import unwrap as _u
from ddtrace.pin import Pin
from ddtrace.vendor.wrapt import wrap_function_wrapper as _w

from .. import trace_utils
from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...ext import net
from ...ext import redis as redisx
from ...internal.utils.formats import stringify_cache_args
from ..redis.util import _trace_redis_cmd
from ..redis.util import _trace_redis_execute_pipeline


try:
    from aioredis.commands.transaction import _RedisBuffer
except ImportError:
    _RedisBuffer = None

config._add("aioredis", dict(_default_service="redis"))

aioredis_version_str = getattr(aioredis, "__version__", "0.0.0")
aioredis_version = tuple([int(i) for i in aioredis_version_str.split(".")])


def patch():
    if getattr(aioredis, "_datadog_patch", False):
        return
    setattr(aioredis, "_datadog_patch", True)
    pin = Pin()
    if aioredis_version >= (2, 0):
        _w("aioredis.client", "Redis.execute_command", traced_execute_command)
        _w("aioredis.client", "Redis.pipeline", traced_pipeline)
        _w("aioredis.client", "Pipeline.execute", traced_execute_pipeline)
        pin.onto(aioredis.client.Redis)
    else:
        _w("aioredis", "Redis.execute", traced_13_execute_command)
        _w("aioredis", "Redis.pipeline", traced_13_pipeline)
        _w("aioredis.commands.transaction", "Pipeline.execute", traced_13_execute_pipeline)
        pin.onto(aioredis.Redis)


def unpatch():
    if not getattr(aioredis, "_datadog_patch", False):
        return

    setattr(aioredis, "_datadog_patch", False)
    if aioredis_version >= (2, 0):
        _u(aioredis.client.Redis, "execute_command")
        _u(aioredis.client.Redis, "pipeline")
        _u(aioredis.client.Pipeline, "execute")
    else:
        _u(aioredis.Redis, "execute")
        _u(aioredis.Redis, "pipeline")
        _u(aioredis.commands.transaction.Pipeline, "execute")


async def traced_execute_command(func, instance, args, kwargs):
    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return await func(*args, **kwargs)

    with _trace_redis_cmd(pin, config.aioredis, instance, args):
        return await func(*args, **kwargs)


def traced_pipeline(func, instance, args, kwargs):
    pipeline = func(*args, **kwargs)
    pin = Pin.get_from(instance)
    if pin:
        pin.onto(pipeline)
    return pipeline


async def traced_execute_pipeline(func, instance, args, kwargs):
    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return await func(*args, **kwargs)

    cmds = [stringify_cache_args(c) for c, _ in instance.command_stack]
    resource = "\n".join(cmds)
    with _trace_redis_execute_pipeline(pin, config.aioredis, resource, instance):
        return await func(*args, **kwargs)


def traced_13_pipeline(func, instance, args, kwargs):
    pipeline = func(*args, **kwargs)
    pin = Pin.get_from(instance)
    if pin:
        pin.onto(pipeline)
    return pipeline


def traced_13_execute_command(func, instance, args, kwargs):
    # If we have a _RedisBuffer then we are in a pipeline
    if isinstance(instance.connection, _RedisBuffer):
        return func(*args, **kwargs)

    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return func(*args, **kwargs)

    # Don't activate the span since this operation is performed as a future which concludes sometime later on in
    # execution so subsequent operations in the stack are not necessarily semantically related
    # (we don't want this span to be the parent of all other spans created before the future is resolved)
    parent = pin.tracer.current_span()
    span = pin.tracer.start_span(
        redisx.CMD,
        service=trace_utils.ext_service(pin, config.aioredis),
        span_type=SpanTypes.REDIS,
        activate=False,
        child_of=parent,
    )

    span.set_tag(SPAN_MEASURED_KEY)
    query = stringify_cache_args(args)
    span.resource = query
    span.set_tag(redisx.RAWCMD, query)
    if pin.tags:
        span.set_tags(pin.tags)

    span.set_tags(
        {
            net.TARGET_HOST: instance.address[0],
            net.TARGET_PORT: instance.address[1],
            redisx.DB: instance.db or 0,
        }
    )
    span.set_metric(redisx.ARGS_LEN, len(args))
    # set analytics sample rate if enabled
    span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.aioredis.get_analytics_sample_rate())

    def _finish_span(future):
        try:
            # Accessing the result will raise an exception if:
            #   - The future was cancelled (CancelledError)
            #   - There was an error executing the future (`future.exception()`)
            #   - The future is in an invalid state
            future.result()
        # CancelledError exceptions extend from BaseException as of Python 3.8, instead of usual Exception
        except BaseException:
            span.set_exc_info(*sys.exc_info())
        finally:
            span.finish()

    task = func(*args, **kwargs)
    # Execute command returns a coroutine when no free connections are available
    # https://github.com/aio-libs/aioredis-py/blob/v1.3.1/aioredis/pool.py#L191
    task = asyncio.ensure_future(task)
    task.add_done_callback(_finish_span)
    return task


async def traced_13_execute_pipeline(func, instance, args, kwargs):
    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return await func(*args, **kwargs)

    cmds = []
    for _, cmd, cmd_args, _ in instance._pipeline:
        parts = [cmd]
        parts.extend(cmd_args)
        cmds.append(stringify_cache_args(parts))
    resource = "\n".join(cmds)
    with pin.tracer.trace(
        redisx.CMD,
        resource=resource,
        service=trace_utils.ext_service(pin, config.aioredis),
        span_type=SpanTypes.REDIS,
    ) as span:

        span.set_tags(
            {
                net.TARGET_HOST: instance._pool_or_conn.address[0],
                net.TARGET_PORT: instance._pool_or_conn.address[1],
                redisx.DB: instance._pool_or_conn.db or 0,
            }
        )

        span.set_tag(SPAN_MEASURED_KEY)
        span.set_tag(redisx.RAWCMD, resource)
        span.set_metric(redisx.PIPELINE_LEN, len(instance._pipeline))
        # set analytics sample rate if enabled
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.aioredis.get_analytics_sample_rate())

        return await func(*args, **kwargs)



import typing

import attr

from ddtrace.profiling import event


@event.event_class
class StackSampleEvent(event.StackBasedEvent):
    """A sample storing executions frames for a thread."""

    # Wall clock
    wall_time_ns = attr.ib(default=0, type=int)
    # CPU time in nanoseconds
    cpu_time_ns = attr.ib(default=0, type=int)


@event.event_class
class StackExceptionSampleEvent(event.StackBasedEvent):
    """A a sample storing raised exceptions and their stack frames."""

    exc_type = attr.ib(default=None, type=typing.Optional[str])



import errno
import json
import os
import os.path
from typing import List
from typing import Set
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union

import attr
from six import ensure_binary

from ddtrace.appsec._ddwaf import DDWaf
from ddtrace.appsec._ddwaf import version
from ddtrace.constants import APPSEC_ENABLED
from ddtrace.constants import APPSEC_EVENT_RULE_ERRORS
from ddtrace.constants import APPSEC_EVENT_RULE_ERROR_COUNT
from ddtrace.constants import APPSEC_EVENT_RULE_LOADED
from ddtrace.constants import APPSEC_EVENT_RULE_VERSION
from ddtrace.constants import APPSEC_JSON
from ddtrace.constants import APPSEC_WAF_DURATION
from ddtrace.constants import APPSEC_WAF_DURATION_EXT
from ddtrace.constants import APPSEC_WAF_VERSION
from ddtrace.constants import MANUAL_KEEP_KEY
from ddtrace.constants import ORIGIN_KEY
from ddtrace.constants import RUNTIME_FAMILY
from ddtrace.contrib.trace_utils import _normalize_tag_name
from ddtrace.ext import SpanTypes
from ddtrace.internal import _context
from ddtrace.internal.logger import get_logger
from ddtrace.internal.processor import SpanProcessor
from ddtrace.internal.rate_limiter import RateLimiter


if TYPE_CHECKING:
    from typing import Dict

    from ddtrace.span import Span

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULT_RULES = os.path.join(ROOT_DIR, "rules.json")
DEFAULT_TRACE_RATE_LIMIT = 100
DEFAULT_WAF_TIMEOUT = 20  # ms
DEFAULT_APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP = (
    r"(?i)(?:p(?:ass)?w(?:or)?d|pass(?:_?phrase)?|secret|(?:api_?|private_?|public_?)key)|token|consumer_?"
    r"(?:id|key|secret)|sign(?:ed|ature)|bearer|authorization"
)
DEFAULT_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP = (
    r"(?i)(?:p(?:ass)?w(?:or)?d|pass(?:_?phrase)?|secret|(?:api_?|private_?|public_?|access_?|secret_?)"
    r"key(?:_?id)?|token|consumer_?(?:id|key|secret)|sign(?:ed|ature)?|auth(?:entication|orization)?)"
    r'(?:\s*=[^;]|"\s*:\s*"[^"]+")|bearer\s+[a-z0-9\._\-]+|token:[a-z0-9]{13}|gh[opsu]_[0-9a-zA-Z]{36}'
    r"|ey[I-L][\w=-]+\.ey[I-L][\w=-]+(?:\.[\w.+\/=-]+)?|[\-]{5}BEGIN[a-z\s]+PRIVATE\sKEY[\-]{5}[^\-]+[\-]"
    r"{5}END[a-z\s]+PRIVATE\sKEY|ssh-rsa\s*[a-z0-9\/\.+]{100,}"
)


log = get_logger(__name__)


def _transform_headers(data):
    # type: (Union[Dict[str, str], List[Tuple[str, str]]]) -> Dict[str, Union[str, List[str]]]
    normalized = {}  # type: Dict[str, Union[str, List[str]]]
    headers = data if isinstance(data, list) else data.items()
    for header, value in headers:
        header = header.lower()
        if header in ("cookie", "set-cookie"):
            continue
        if header in normalized:  # if a header with the same lowercase name already exists, let's make it an array
            existing = normalized[header]
            if isinstance(existing, list):
                existing.append(value)
            else:
                normalized[header] = [existing, value]
        else:
            normalized[header] = value
    return normalized


def get_rules():
    # type: () -> str
    return os.getenv("DD_APPSEC_RULES", default=DEFAULT_RULES)


def get_appsec_obfuscation_parameter_key_regexp():
    # type: () -> bytes
    return ensure_binary(
        os.getenv("DD_APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP", DEFAULT_APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP)
    )


def get_appsec_obfuscation_parameter_value_regexp():
    # type: () -> bytes
    return ensure_binary(
        os.getenv("DD_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP", DEFAULT_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP)
    )


class _Addresses(object):
    SERVER_REQUEST_BODY = "server.request.body"
    SERVER_REQUEST_QUERY = "server.request.query"
    SERVER_REQUEST_HEADERS_NO_COOKIES = "server.request.headers.no_cookies"
    SERVER_REQUEST_URI_RAW = "server.request.uri.raw"
    SERVER_REQUEST_METHOD = "server.request.method"
    SERVER_REQUEST_PATH_PARAMS = "server.request.path_params"
    SERVER_REQUEST_COOKIES = "server.request.cookies"
    SERVER_RESPONSE_STATUS = "server.response.status"
    SERVER_RESPONSE_HEADERS_NO_COOKIES = "server.response.headers.no_cookies"


_COLLECTED_REQUEST_HEADERS = {
    "accept",
    "accept-encoding",
    "accept-language",
    "content-encoding",
    "content-language",
    "content-length",
    "content-type",
    "forwarded",
    "forwarded-for",
    "host",
    "true-client-ip",
    "user-agent",
    "via",
    "x-client-ip",
    "x-cluster-client-ip",
    "x-forwarded",
    "x-forwarded-for",
    "x-real-ip",
}

_COLLECTED_HEADER_PREFIX = "http.request.headers."


def _set_headers(span, headers, kind):
    # type: (Span, Dict[str, Union[str, List[str]]], str) -> None
    for k in headers:
        if k.lower() in _COLLECTED_REQUEST_HEADERS:
            # since the header value can be a list, use `set_tag()` to ensure it is converted to a string
            span.set_tag(_normalize_tag_name(kind, k), headers[k])


def _get_rate_limiter():
    # type: () -> RateLimiter
    return RateLimiter(int(os.getenv("DD_APPSEC_TRACE_RATE_LIMIT", DEFAULT_TRACE_RATE_LIMIT)))


def _get_waf_timeout():
    # type: () -> int
    return int(os.getenv("DD_APPSEC_WAF_TIMEOUT", DEFAULT_WAF_TIMEOUT))


@attr.s(eq=False)
class AppSecSpanProcessor(SpanProcessor):

    rules = attr.ib(type=str, factory=get_rules)
    obfuscation_parameter_key_regexp = attr.ib(type=bytes, factory=get_appsec_obfuscation_parameter_key_regexp)
    obfuscation_parameter_value_regexp = attr.ib(type=bytes, factory=get_appsec_obfuscation_parameter_value_regexp)
    _ddwaf = attr.ib(type=DDWaf, default=None)
    _addresses_to_keep = attr.ib(type=Set[str], factory=set)
    _rate_limiter = attr.ib(type=RateLimiter, factory=_get_rate_limiter)
    _waf_timeout = attr.ib(type=int, factory=_get_waf_timeout)

    @property
    def enabled(self):
        return self._ddwaf is not None

    def __attrs_post_init__(self):
        # type: () -> None
        if self._ddwaf is None:
            try:
                with open(self.rules, "r") as f:
                    rules = json.load(f)
            except EnvironmentError as err:
                if err.errno == errno.ENOENT:
                    log.error(
                        "[DDAS-0001-03] AppSec could not read the rule file %s. Reason: file does not exist", self.rules
                    )
                else:
                    # TODO: try to log reasons
                    log.error("[DDAS-0001-03] AppSec could not read the rule file %s.", self.rules)
                raise
            except json.decoder.JSONDecodeError:
                log.error(
                    "[DDAS-0001-03] AppSec could not read the rule file %s. Reason: invalid JSON file", self.rules
                )
                raise
            except Exception:
                # TODO: try to log reasons
                log.error("[DDAS-0001-03] AppSec could not read the rule file %s.", self.rules)
                raise
            try:
                self._ddwaf = DDWaf(
                    rules, self.obfuscation_parameter_key_regexp, self.obfuscation_parameter_value_regexp
                )
            except ValueError:
                # Partial of DDAS-0005-00
                log.warning("[DDAS-0005-00] WAF initialization failed")
                raise
        for address in self._ddwaf.required_data:
            self._mark_needed(address)
        # we always need the request headers
        self._mark_needed(_Addresses.SERVER_REQUEST_HEADERS_NO_COOKIES)
        # we always need the response headers
        self._mark_needed(_Addresses.SERVER_RESPONSE_HEADERS_NO_COOKIES)

    def on_span_start(self, span):
        # type: (Span) -> None
        pass

    def _mark_needed(self, address):
        # type: (str) -> None
        self._addresses_to_keep.add(address)

    def _is_needed(self, address):
        # type: (str) -> bool
        return address in self._addresses_to_keep

    def on_span_finish(self, span):
        # type: (Span) -> None
        if span.span_type != SpanTypes.WEB:
            return
        span.set_metric(APPSEC_ENABLED, 1.0)
        span._set_str_tag(RUNTIME_FAMILY, "python")

        data = {}
        if self._is_needed(_Addresses.SERVER_REQUEST_QUERY):
            request_query = _context.get_item("http.request.query", span=span)
            if request_query is not None:
                data[_Addresses.SERVER_REQUEST_QUERY] = request_query

        if self._is_needed(_Addresses.SERVER_REQUEST_HEADERS_NO_COOKIES):
            request_headers = _context.get_item("http.request.headers", span=span)
            if request_headers is not None:
                data[_Addresses.SERVER_REQUEST_HEADERS_NO_COOKIES] = _transform_headers(request_headers)

        if self._is_needed(_Addresses.SERVER_REQUEST_URI_RAW):
            uri = _context.get_item("http.request.uri", span=span)
            if uri is not None:
                data[_Addresses.SERVER_REQUEST_URI_RAW] = uri

        if self._is_needed(_Addresses.SERVER_REQUEST_METHOD):
            request_method = _context.get_item("http.request.method", span=span)
            if request_method is not None:
                data[_Addresses.SERVER_REQUEST_METHOD] = request_method

        if self._is_needed(_Addresses.SERVER_REQUEST_PATH_PARAMS):
            path_params = _context.get_item("http.request.path_params", span=span)
            if path_params is not None:
                data[_Addresses.SERVER_REQUEST_PATH_PARAMS] = path_params

        if self._is_needed(_Addresses.SERVER_REQUEST_COOKIES):
            cookies = _context.get_item("http.request.cookies", span=span)
            if cookies is not None:
                data[_Addresses.SERVER_REQUEST_COOKIES] = cookies

        if self._is_needed(_Addresses.SERVER_RESPONSE_STATUS):
            status = _context.get_item("http.response.status", span=span)
            if status is not None:
                data[_Addresses.SERVER_RESPONSE_STATUS] = status

        if self._is_needed(_Addresses.SERVER_RESPONSE_HEADERS_NO_COOKIES):
            response_headers = _context.get_item("http.response.headers", span=span)
            if response_headers is not None:
                data[_Addresses.SERVER_RESPONSE_HEADERS_NO_COOKIES] = _transform_headers(response_headers)

        if self._is_needed(_Addresses.SERVER_REQUEST_BODY):
            body = _context.get_item("http.request.body", span=span)
            if body is not None:
                data[_Addresses.SERVER_REQUEST_BODY] = body

        log.debug("[DDAS-001-00] Executing AppSec In-App WAF with parameters: %s", data)
        res, total_runtime, total_overall_runtime = self._ddwaf.run(data, self._waf_timeout)  # res is a serialized json

        try:
            info = self._ddwaf.info
            if info["errors"]:
                span._set_str_tag(APPSEC_EVENT_RULE_ERRORS, json.dumps(info["errors"]))
            span._set_str_tag(APPSEC_EVENT_RULE_VERSION, info["version"])
            span._set_str_tag(APPSEC_WAF_VERSION, "%s.%s.%s" % version())

            span.set_metric(APPSEC_EVENT_RULE_LOADED, info["loaded"])
            span.set_metric(APPSEC_EVENT_RULE_ERROR_COUNT, info["failed"])
            span.set_metric(APPSEC_WAF_DURATION, total_runtime)
            span.set_metric(APPSEC_WAF_DURATION_EXT, total_overall_runtime)
        except (json.decoder.JSONDecodeError, ValueError):
            log.warning("Error parsing data AppSec In-App WAF metrics report")
        except Exception:
            log.warning("Error executing AppSec In-App WAF metrics report: %s", exc_info=True)
        if res is not None:
            # We run the rate limiter only if there is an attack, its goal is to limit the number of collected asm
            # events
            allowed = self._rate_limiter.is_allowed(span.start_ns)
            if not allowed:
                # TODO: add metric collection to keep an eye (when it's name is clarified)
                return
            if _Addresses.SERVER_REQUEST_HEADERS_NO_COOKIES in data:
                _set_headers(span, data[_Addresses.SERVER_REQUEST_HEADERS_NO_COOKIES], kind="request")

            if _Addresses.SERVER_RESPONSE_HEADERS_NO_COOKIES in data:
                _set_headers(span, data[_Addresses.SERVER_RESPONSE_HEADERS_NO_COOKIES], kind="response")
            # Partial DDAS-011-00
            log.debug("[DDAS-011-00] AppSec In-App WAF returned: %s", res)
            span._set_str_tag("appsec.event", "true")
            span._set_str_tag(APPSEC_JSON, '{"triggers":%s}' % (res,))

            remote_ip = _context.get_item("http.request.remote_ip", span=span)
            if remote_ip:
                # Note that if the ip collection is disabled by the env var
                # DD_TRACE_CLIENT_IP_HEADER_DISABLED actor.ip won't be sent
                span._set_str_tag("actor.ip", remote_ip)
            # Right now, we overwrite any value that could be already there. We need to reconsider when ASM/AppSec's
            # specs are updated.
            span.set_tag(MANUAL_KEEP_KEY)
            if span.get_tag(ORIGIN_KEY) is None:
                span._set_str_tag(ORIGIN_KEY, "appsec")



import os
import time
from typing import Dict
from typing import List
from typing import Optional

from ...internal import atexit
from ...internal import forksafe
from ...settings import _config as config
from ..agent import get_connection
from ..agent import get_trace_url
from ..compat import get_connection_response
from ..compat import httplib
from ..encoding import JSONEncoderV2
from ..logger import get_logger
from ..periodic import PeriodicService
from ..runtime import get_runtime_id
from ..service import ServiceStatus
from ..utils.formats import parse_tags_str
from ..utils.time import StopWatch
from .data import get_application
from .data import get_dependencies
from .data import get_host_info


log = get_logger(__name__)


def _get_interval_or_default():
    return float(os.getenv("DD_INSTRUMENTATION_TELEMETRY_INTERVAL_SECONDS", default=60))


class TelemetryWriter(PeriodicService):
    """
    Periodic service which sends Telemetry request payloads to the agent.
    Supports version one of the instrumentation telemetry api
    """

    # telemetry endpoint uses events platform v2 api
    ENDPOINT = "telemetry/proxy/api/v2/apmtelemetry"
    HEARTBEAT_MIN_INTERVAL = 60  # type: int

    def __init__(self, agent_url=None):
        # type: (Optional[str]) -> None
        super(TelemetryWriter, self).__init__(interval=_get_interval_or_default())

        # _enabled is None at startup, and is only set to true or false
        # after the config has been processed
        self._enabled = None  # type: Optional[bool]
        self._agent_url = agent_url or get_trace_url()

        self._encoder = JSONEncoderV2()
        self._events_queue = []  # type: List[Dict]
        self._integrations_queue = []  # type: List[Dict]
        self._lock = forksafe.Lock()  # type: forksafe.ResetObject
        self._forked = False  # type: bool
        # Set initial heartbeat time to now since we'll be sending an app-started event
        self._last_heartbeat = time.time()  # type: float

        self._headers = {
            "Content-type": "application/json",
            "DD-Telemetry-API-Version": "v1",
        }  # type: Dict[str, str]
        additional_header_str = os.environ.get("_DD_TELEMETRY_WRITER_ADDITIONAL_HEADERS")
        if additional_header_str is not None:
            self._headers.update(parse_tags_str(additional_header_str))

        # _sequence is a counter representing the number of requests sent by the writer
        self._sequence = 1  # type: int

    @property
    def url(self):
        return "%s/%s" % (self._agent_url, self.ENDPOINT)

    def _send_request(self, request):
        # type: (Dict) -> httplib.HTTPResponse
        """Sends a telemetry request to the trace agent"""
        with StopWatch() as sw:
            try:
                conn = get_connection(self._agent_url)
                rb_json = self._encoder.encode(request)
                conn.request("POST", self.ENDPOINT, rb_json, self._create_headers(request["request_type"]))

                resp = get_connection_response(conn)
                log.debug(
                    "sent %d in %.5fs to %s/%s. response: %s",
                    len(rb_json),
                    sw.elapsed(),
                    self._agent_url,
                    self.ENDPOINT,
                    resp.status,
                )
                return resp
            finally:
                conn.close()

    def _flush_integrations_queue(self):
        # type: () -> List[Dict]
        """Returns a list of all integrations queued by add_integration"""
        with self._lock:
            integrations = self._integrations_queue
            self._integrations_queue = []
        return integrations

    def _flush_events_queue(self):
        # type: () -> List[Dict]
        """Returns a list of all integrations queued by classmethods"""
        with self._lock:
            requests = self._events_queue
            self._events_queue = []
        return requests

    def reset_queues(self):
        # type: () -> None
        with self._lock:
            self._integrations_queue = []
            self._events_queue = []

    def periodic(self):
        integrations = self._flush_integrations_queue()
        if integrations:
            self._app_integrations_changed_event(integrations)

        if not self._events_queue:
            # Optimization: only queue heartbeat if no other events are queued
            self.app_heartbeat_event()

        telemetry_requests = self._flush_events_queue()

        for telemetry_request in telemetry_requests:
            try:
                resp = self._send_request(telemetry_request)
                if resp.status >= 300:
                    log.debug(
                        "failed to send telemetry to the Datadog Agent at %s/%s. response: %s",
                        self._agent_url,
                        self.ENDPOINT,
                        resp.status,
                    )
            except Exception:
                log.debug(
                    "failed to send telemetry to the Datadog Agent at %s/%s.",
                    self._agent_url,
                    self.ENDPOINT,
                    exc_info=True,
                )

    def _start_service(self, *args, **kwargs):
        # type: (...) -> None
        self.app_started_event()
        return super(TelemetryWriter, self)._start_service(*args, **kwargs)

    def on_shutdown(self):
        self._app_closing_event()
        self.periodic()

    def _stop_service(self, *args, **kwargs):
        # type: (...) -> None
        super(TelemetryWriter, self)._stop_service(*args, **kwargs)
        self.join()

    def add_event(self, payload, payload_type):
        # type: (Dict, str) -> None
        """
        Adds a Telemetry Request to the TelemetryWriter request buffer

        :param Dict payload: stores a formatted telemetry request
        :param str payload_type: The payload_type denotes the type of telmetery request.
            Payload types accepted by telemetry/proxy v1: app-started, app-closing, app-integrations-change
        """
        with self._lock:
            if not self._enabled:
                return

            request = self._create_telemetry_request(payload, payload_type, self._sequence)
            self._sequence += 1
            self._events_queue.append(request)

    def add_integration(self, integration_name, auto_enabled):
        # type: (str, bool) -> None
        """
        Creates and queues the names and settings of a patched module

        :param str integration_name: name of patched module
        :param bool auto_enabled: True if module is enabled in _monkey.PATCH_MODULES
        """
        with self._lock:
            if self._enabled is not None and not self._enabled:
                return

            integration = {
                "name": integration_name,
                "version": "",
                "enabled": True,
                "auto_enabled": auto_enabled,
                "compatible": True,
                "error": "",
            }
            self._integrations_queue.append(integration)

    def app_started_event(self):
        # type: () -> None
        """Sent when TelemetryWriter is enabled or forks"""
        if self._forked:
            # app-started events should only be sent by the main process
            return
        payload = {
            "dependencies": get_dependencies(),
            "integrations": self._flush_integrations_queue(),
            "configurations": [],
        }
        self.add_event(payload, "app-started")

    def app_heartbeat_event(self):
        # type: () -> None
        if self._forked:
            # TODO: Enable app-heartbeat on forks
            #   Since we only send app-started events in the main process
            #   any forked processes won't be able to access the list of
            #   dependencies for this app, and therefore app-heartbeat won't
            #   add much value today.
            return

        # DEV: Although the default flush interval is 60 seconds,
        #   we want to explicitly control the flow here in case
        #   the flush interval ever changes.
        #
        #   If the flush interval ever exceeds 60 seconds, then we
        #   will queue a heartbeat event every flush:
        #   60 <= X <= flush interval.
        now = time.time()
        if now - self._last_heartbeat >= TelemetryWriter.HEARTBEAT_MIN_INTERVAL:
            self.add_event({}, "app-heartbeat")
            self._last_heartbeat = now

    def _app_closing_event(self):
        # type: () -> None
        """Adds a Telemetry request which notifies the agent that an application instance has terminated"""
        if self._forked:
            # app-closing event should only be sent by the main process
            return
        payload = {}  # type: Dict
        self.add_event(payload, "app-closing")

    def _app_integrations_changed_event(self, integrations):
        # type: (List[Dict]) -> None
        """Adds a Telemetry request which sends a list of configured integrations to the agent"""
        payload = {
            "integrations": integrations,
        }
        self.add_event(payload, "app-integrations-change")

    def _create_headers(self, payload_type):
        # type: (str) -> Dict
        """Creates request headers"""
        headers = self._headers.copy()
        headers["DD-Telemetry-Request-Type"] = payload_type
        return headers

    def _create_telemetry_request(self, payload, payload_type, sequence_id):
        # type: (Dict, str, int) -> Dict
        """Initializes the required fields for a generic Telemetry Intake Request"""
        return {
            "tracer_time": int(time.time()),
            "runtime_id": get_runtime_id(),
            "api_version": "v1",
            "seq_id": sequence_id,
            "application": get_application(config.service, config.version, config.env),
            "host": get_host_info(),
            "payload": payload,
            "request_type": payload_type,
        }

    def _fork_writer(self):
        # type: () -> None
        self._forked = True
        # Avoid sending duplicate events.
        # Queued events should be sent in the main process.
        self.reset_queues()

    def disable(self):
        # type: () -> None
        """
        Disable the telemetry collection service and drop the existing integrations and events
        Once disabled, telemetry collection can be re-enabled by calling ``enable`` again.
        """
        with self._lock:
            self._enabled = False
        self.reset_queues()
        if self.status == ServiceStatus.STOPPED:
            return

        forksafe.unregister(self._fork_writer)
        atexit.unregister(self.stop)

        self.stop()

    def enable(self):
        # type: () -> None
        """
        Enable the instrumentation telemetry collection service. If the service has already been
        activated before, this method does nothing. Use ``disable`` to turn off the telemetry collection service.
        """
        if self.status == ServiceStatus.RUNNING:
            return

        self._enabled = True
        self.start()

        forksafe.register(self._fork_writer)
        atexit.register(self.stop)



import asyncio

import sanic

import ddtrace
from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.ext import SpanTypes
from ddtrace.internal.utils.wrappers import unwrap as _u
from ddtrace.pin import Pin
from ddtrace.vendor import wrapt
from ddtrace.vendor.wrapt import wrap_function_wrapper as _w

from .. import trace_utils
from ...internal.logger import get_logger


log = get_logger(__name__)

config._add("sanic", dict(_default_service="sanic", distributed_tracing=True))

SANIC_VERSION = (0, 0, 0)


def _get_current_span(request):
    pin = Pin._find(request.ctx)
    if not pin or not pin.enabled():
        return None

    return pin.tracer.current_span()


def update_span(span, response):
    # Check for response status or headers on the response object
    # DEV: This object can either be a form of BaseResponse or an Exception
    #      if we do not have a status code, we can assume this is an exception
    #      and so use 500
    status_code = getattr(response, "status", 500)
    response_headers = getattr(response, "headers", None)
    trace_utils.set_http_meta(span, config.sanic, status_code=status_code, response_headers=response_headers)


def _wrap_response_callback(span, callback):
    # Only for sanic 20 and older
    # Wrap response callbacks (either sync or async function) to set HTTP
    # response span tags

    @wrapt.function_wrapper
    def wrap_sync(wrapped, instance, args, kwargs):
        r = wrapped(*args, **kwargs)
        response = args[0]
        update_span(span, response)
        return r

    @wrapt.function_wrapper
    async def wrap_async(wrapped, instance, args, kwargs):
        r = await wrapped(*args, **kwargs)
        response = args[0]
        update_span(span, response)
        return r

    if asyncio.iscoroutinefunction(callback):
        return wrap_async(callback)

    return wrap_sync(callback)


async def patch_request_respond(wrapped, instance, args, kwargs):
    # Only for sanic 21 and newer
    # Wrap the framework response to set HTTP response span tags
    response = await wrapped(*args, **kwargs)
    span = _get_current_span(instance)
    if not span:
        return response

    update_span(span, response)

    # Sanic 21.9.x does not dispatch `http.lifecycle.response` in `handle_exception`
    #  so we have to handle finishing the span here instead
    if (21, 9, 0) <= SANIC_VERSION < (21, 12, 0) and getattr(instance.ctx, "__dd_span_call_finish", False):
        span.finish()
    return response


def _get_path(request):
    """Get path and replace path parameter values with names if route exists."""
    path = request.path
    try:
        match_info = request.match_info
    except sanic.exceptions.SanicException:
        return path
    for key, value in match_info.items():
        try:
            value = str(value)
        except Exception:
            # Best effort
            continue
        path = path.replace(value, f"<{key}>")
    return path


async def patch_run_request_middleware(wrapped, instance, args, kwargs):
    # Set span resource from the framework request
    request = args[0]
    span = _get_current_span(request)
    if span is not None:
        span.resource = "{} {}".format(request.method, _get_path(request))
    return await wrapped(*args, **kwargs)


def patch():
    """Patch the instrumented methods."""
    global SANIC_VERSION

    if getattr(sanic, "__datadog_patch", False):
        return
    setattr(sanic, "__datadog_patch", True)

    SANIC_VERSION = tuple(map(int, sanic.__version__.split(".")))

    if SANIC_VERSION >= (21, 9, 0):
        _w("sanic", "Sanic.__init__", patch_sanic_init)
        _w(sanic.request, "Request.respond", patch_request_respond)
    else:
        _w("sanic", "Sanic.handle_request", patch_handle_request)
        if SANIC_VERSION >= (21, 0, 0):
            _w("sanic", "Sanic._run_request_middleware", patch_run_request_middleware)
            _w(sanic.request, "Request.respond", patch_request_respond)


def unpatch():
    """Unpatch the instrumented methods."""
    if not getattr(sanic, "__datadog_patch", False):
        return

    if SANIC_VERSION >= (21, 9, 0):
        _u(sanic.Sanic, "__init__")
        _u(sanic.request.Request, "respond")
    else:
        _u(sanic.Sanic, "handle_request")
        if SANIC_VERSION >= (21, 0, 0):
            _u(sanic.Sanic, "_run_request_middleware")
            _u(sanic.request.Request, "respond")

    setattr(sanic, "__datadog_patch", False)


def patch_sanic_init(wrapped, instance, args, kwargs):
    """Wrapper for creating sanic apps to automatically add our signal handlers"""
    wrapped(*args, **kwargs)

    instance.add_signal(sanic_http_lifecycle_handle, "http.lifecycle.handle")
    instance.add_signal(sanic_http_routing_after, "http.routing.after")
    instance.add_signal(sanic_http_lifecycle_exception, "http.lifecycle.exception")
    instance.add_signal(sanic_http_lifecycle_response, "http.lifecycle.response")


async def patch_handle_request(wrapped, instance, args, kwargs):
    """Wrapper for Sanic.handle_request"""

    def unwrap(request, write_callback=None, stream_callback=None, **kwargs):
        return request, write_callback, stream_callback, kwargs

    request, write_callback, stream_callback, new_kwargs = unwrap(*args, **kwargs)

    if request.scheme not in ("http", "https"):
        return await wrapped(*args, **kwargs)

    with _create_sanic_request_span(request) as span:
        if write_callback is not None:
            new_kwargs["write_callback"] = _wrap_response_callback(span, write_callback)
        if stream_callback is not None:
            new_kwargs["stream_callback"] = _wrap_response_callback(span, stream_callback)

        return await wrapped(request, **new_kwargs)


def _create_sanic_request_span(request):
    """Helper to create sanic.request span and attach a pin to request.ctx"""
    pin = Pin()
    pin.onto(request.ctx)

    if SANIC_VERSION < (21, 0, 0):
        # Set span resource from the framework request
        resource = "{} {}".format(request.method, _get_path(request))
    else:
        # The path is not available anymore in 21.x. Get it from
        # the _run_request_middleware instrumented method.
        resource = None

    headers = request.headers.copy()

    trace_utils.activate_distributed_headers(ddtrace.tracer, int_config=config.sanic, request_headers=headers)

    span = pin.tracer.trace(
        "sanic.request",
        service=trace_utils.int_service(None, config.sanic),
        resource=resource,
        span_type=SpanTypes.WEB,
    )
    sample_rate = config.sanic.get_analytics_sample_rate(use_global_config=True)
    if sample_rate is not None:
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, sample_rate)

    method = request.method
    url = "{scheme}://{host}{path}".format(scheme=request.scheme, host=request.host, path=request.path)
    query_string = request.query_string
    if isinstance(query_string, bytes):
        query_string = query_string.decode()
    trace_utils.set_http_meta(span, config.sanic, method=method, url=url, query=query_string, request_headers=headers)

    return span


async def sanic_http_lifecycle_handle(request):
    """Lifecycle signal called when a new request is started."""
    _create_sanic_request_span(request)


async def sanic_http_routing_after(request, route, kwargs, handler):
    """Lifecycle signal called after routing has been resolved."""
    span = _get_current_span(request)
    if not span:
        return

    pattern = route.raw_path
    # Sanic 21.9.0 and newer strip the leading slash from `route.raw_path`
    if not pattern.startswith("/"):
        pattern = "/{}".format(pattern)
    if route.regex:
        pattern = route.pattern

    span.resource = "{} {}".format(request.method, pattern)
    span.set_tag("sanic.route.name", route.name)


async def sanic_http_lifecycle_response(request, response):
    """Lifecycle signal called when a response is starting.

    Note: This signal does not get called when exceptions occur
          in 21.9.x. The issue was resolved in 21.12.x
    """
    span = _get_current_span(request)
    if not span:
        return
    try:
        update_span(span, response)
    finally:
        span.finish()


async def sanic_http_lifecycle_exception(request, exception):
    """Lifecycle signal called when an exception occurs."""
    span = _get_current_span(request)
    if not span:
        return

    # Do not attach exception for exceptions not considered as errors
    # ex: Http 400s
    # DEV: We still need to set `__dd_span_call_finish` below
    if not hasattr(exception, "status_code") or config.http_server.is_error_code(exception.status_code):
        ex_type = type(exception)
        ex_tb = getattr(exception, "__traceback__", None)
        span.set_exc_info(ex_type, exception, ex_tb)

    # Sanic 21.9.x does not dispatch `http.lifecycle.response` in `handle_exception`
    #  so we need to indicate to `patch_request_respond` to finish the span
    if (21, 9, 0) <= SANIC_VERSION < (21, 12, 0):
        request.ctx.__dd_span_call_finish = True



# -*- encoding: utf-8 -*-
import multiprocessing
import os
import sys

import pytest
import six

from tests.utils import call_program

from . import utils


def test_call_script(monkeypatch):
    # Set a very short timeout to exit fast
    monkeypatch.setenv("DD_PROFILING_API_TIMEOUT", "0.1")
    monkeypatch.setenv("DD_PROFILING_ENABLED", "1")
    stdout, stderr, exitcode, _ = call_program(
        "ddtrace-run", sys.executable, os.path.join(os.path.dirname(__file__), "simple_program.py")
    )
    if sys.platform == "win32":
        assert exitcode == 0, (stdout, stderr)
    else:
        assert exitcode == 42, (stdout, stderr)
    hello, interval, stacks, pid = list(s.strip() for s in stdout.decode().strip().split("\n"))
    assert hello == "hello world"
    assert float(interval) >= 0.01
    assert int(stacks) >= 1


@pytest.mark.skipif(not os.getenv("DD_PROFILE_TEST_GEVENT", False), reason="Not testing gevent")
def test_call_script_gevent(monkeypatch):
    monkeypatch.setenv("DD_PROFILING_API_TIMEOUT", "0.1")
    stdout, stderr, exitcode, pid = call_program(
        sys.executable, os.path.join(os.path.dirname(__file__), "simple_program_gevent.py")
    )
    assert exitcode == 0, (stdout, stderr)


def test_call_script_pprof_output(tmp_path, monkeypatch):
    """This checks if the pprof output and atexit register work correctly.

    The script does not run for one minute, so if the `stop_on_exit` flag is broken, this test will fail.
    """
    filename = str(tmp_path / "pprof")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", filename)
    monkeypatch.setenv("DD_PROFILING_CAPTURE_PCT", "1")
    monkeypatch.setenv("DD_PROFILING_ENABLED", "1")
    stdout, stderr, exitcode, _ = call_program(
        "ddtrace-run", sys.executable, os.path.join(os.path.dirname(__file__), "simple_program.py")
    )
    if sys.platform == "win32":
        assert exitcode == 0, (stdout, stderr)
    else:
        assert exitcode == 42, (stdout, stderr)
    hello, interval, stacks, pid = list(s.strip() for s in stdout.decode().strip().split("\n"))
    utils.check_pprof_file(filename + "." + str(pid) + ".1")
    return filename, pid


@pytest.mark.skipif(six.PY2, reason="This test deadlocks randomly on Python 2")
@pytest.mark.skipif(sys.platform == "win32", reason="fork only available on Unix")
def test_fork(tmp_path, monkeypatch):
    filename = str(tmp_path / "pprof")
    monkeypatch.setenv("DD_PROFILING_API_TIMEOUT", "0.1")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", filename)
    monkeypatch.setenv("DD_PROFILING_CAPTURE_PCT", "100")
    stdout, stderr, exitcode, pid = call_program(
        "python", os.path.join(os.path.dirname(__file__), "simple_program_fork.py")
    )
    assert exitcode == 0
    child_pid = stdout.decode().strip()
    utils.check_pprof_file(filename + "." + str(pid) + ".1")
    utils.check_pprof_file(filename + "." + str(child_pid) + ".1")


@pytest.mark.skipif(sys.platform == "win32", reason="fork only available on Unix")
@pytest.mark.skipif(not os.getenv("DD_PROFILE_TEST_GEVENT", False), reason="Not testing gevent")
def test_fork_gevent(monkeypatch):
    monkeypatch.setenv("DD_PROFILING_API_TIMEOUT", "0.1")
    stdout, stderr, exitcode, pid = call_program("python", os.path.join(os.path.dirname(__file__), "gevent_fork.py"))
    assert exitcode == 0


atleast_py37 = sys.version_info[:2] >= (3, 7)

if atleast_py37:
    methods = multiprocessing.get_all_start_methods()
else:
    methods = []


@pytest.mark.parametrize(
    "method",
    methods,
)
def test_multiprocessing(method, tmp_path, monkeypatch):
    filename = str(tmp_path / "pprof")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", filename)
    monkeypatch.setenv("DD_PROFILING_ENABLED", "1")
    monkeypatch.setenv("DD_PROFILING_CAPTURE_PCT", "1")
    monkeypatch.setenv("DD_PROFILING_UPLOAD_INTERVAL", "0.1")
    stdout, stderr, exitcode, _ = call_program(
        "ddtrace-run",
        sys.executable,
        os.path.join(os.path.dirname(__file__), "_test_multiprocessing.py"),
        method,
    )
    assert exitcode == 0, (stdout, stderr)
    pid, child_pid = list(s.strip() for s in stdout.decode().strip().split("\n"))
    utils.check_pprof_file(filename + "." + str(pid) + ".1")
    utils.check_pprof_file(filename + "." + str(child_pid) + ".1")



#!/usr/bin/env python
# coding: utf-8

import datetime

from msgpack import FormatError
from msgpack import OutOfData
from msgpack import StackError
from msgpack import Unpacker
from msgpack import unpackb
from pytest import raises

from ddtrace.internal._encoding import packb


class DummyException(Exception):
    pass


def test_raise_on_find_unsupported_value():
    with raises(TypeError):
        packb(datetime.datetime.now())


def test_raise_from_object_hook():
    def hook(obj):
        raise DummyException

    raises(DummyException, unpackb, packb({}), object_hook=hook)
    raises(DummyException, unpackb, packb({"fizz": "buzz"}), object_hook=hook)
    raises(DummyException, unpackb, packb({"fizz": "buzz"}), object_pairs_hook=hook)
    raises(DummyException, unpackb, packb({"fizz": {"buzz": "spam"}}), object_hook=hook)
    raises(
        DummyException,
        unpackb,
        packb({"fizz": {"buzz": "spam"}}),
        object_pairs_hook=hook,
    )


def test_invalidvalue():
    incomplete = b"\xd9\x97#DL_"  # raw8 - length=0x97
    with raises(ValueError):
        unpackb(incomplete)

    with raises(OutOfData):
        unpacker = Unpacker()
        unpacker.feed(incomplete)
        unpacker.unpack()

    with raises(FormatError):
        unpackb(b"\xc1")  # (undefined tag)

    with raises(FormatError):
        unpackb(b"\x91\xc1")  # fixarray(len=1) [ (undefined tag) ]

    with raises(StackError):
        unpackb(b"\x91" * 3000)  # nested fixarray(len=1)



import opentracing
from opentracing import Format

import ddtrace
from ddtrace.opentracer.span_context import SpanContext


class TestTracerCompatibility(object):
    """Ensure that our opentracer produces results in the underlying ddtracer."""

    def test_ottracer_uses_global_ddtracer(self):
        """Ensure that the opentracer will by default use the global ddtracer
        as its underlying Datadog tracer.
        """
        tracer = ddtrace.opentracer.Tracer()
        assert tracer._dd_tracer is ddtrace.tracer

    def test_custom_ddtracer(self):
        """A user should be able to specify their own Datadog tracer instance if
        they wish.
        """
        custom_dd_tracer = ddtrace.Tracer()
        tracer = ddtrace.opentracer.Tracer(dd_tracer=custom_dd_tracer)
        assert tracer._dd_tracer is custom_dd_tracer

    def test_ot_dd_global_tracers(self, global_tracer):
        """Ensure our test function opentracer_init() prep"""
        ot_tracer = global_tracer
        dd_tracer = global_tracer._dd_tracer

        # check all the global references
        assert ot_tracer is opentracing.tracer
        assert ot_tracer._dd_tracer is dd_tracer
        assert dd_tracer is ddtrace.tracer

    def test_ot_dd_nested_trace(self, ot_tracer, dd_tracer, test_spans):
        """Ensure intertwined usage of the opentracer and ddtracer."""

        with ot_tracer.start_span("my_ot_span") as ot_span:
            with dd_tracer.trace("my_dd_span") as dd_span:
                pass
        spans = test_spans.pop()
        assert len(spans) == 2

        # confirm the ordering
        assert spans[1] is ot_span._dd_span
        assert spans[0] is dd_span

        # check the parenting
        assert spans[0].parent_id is None
        assert spans[1].parent_id is None

    def test_dd_ot_nested_trace(self, ot_tracer, dd_tracer, test_spans):
        """Ensure intertwined usage of the opentracer and ddtracer."""
        with dd_tracer.trace("my_dd_span") as dd_span:
            with ot_tracer.start_span("my_ot_span") as ot_span:
                pass
        spans = test_spans.pop()
        assert len(spans) == 2

        # confirm the ordering
        assert spans[0] is dd_span
        assert spans[1] is ot_span._dd_span

        # check the parenting
        assert spans[0].parent_id is None
        assert spans[1].parent_id is spans[0].span_id

    def test_ot_dd_ot_dd_nested_trace(self, ot_tracer, dd_tracer, test_spans):
        """Ensure intertwined usage of the opentracer and ddtracer."""
        with ot_tracer.start_active_span("ot_span") as ot_scope:
            with dd_tracer.trace("dd_span") as dd_span:
                with ot_tracer.start_active_span("ot_span2") as ot_scope2:
                    with dd_tracer.trace("dd_span2") as dd_span2:
                        pass

        spans = test_spans.pop()
        assert len(spans) == 4

        spans = {span.name: span for span in spans}
        assert spans["ot_span"] == ot_scope.span._dd_span
        assert spans["dd_span"] == dd_span
        assert spans["ot_span2"] == ot_scope2.span._dd_span
        assert spans["dd_span2"] == dd_span2

        # check the parenting
        assert spans["ot_span"].parent_id is None
        assert spans["dd_span"].parent_id is spans["ot_span"].span_id
        assert spans["ot_span2"].parent_id is spans["dd_span"].span_id
        assert spans["dd_span2"].parent_id is spans["ot_span2"].span_id

    def test_ot_ot_dd_ot_dd_nested_trace_active(self, ot_tracer, dd_tracer, test_spans):
        """Ensure intertwined usage of the opentracer and ddtracer."""
        with ot_tracer.start_active_span("my_ot_span") as ot_scope:
            with ot_tracer.start_active_span("my_ot_span") as ot_scope2:
                with dd_tracer.trace("my_dd_span") as dd_span:
                    with ot_tracer.start_active_span("my_ot_span") as ot_scope3:
                        with dd_tracer.trace("my_dd_span") as dd_span2:
                            pass

        spans = test_spans.pop()
        assert len(spans) == 5

        # confirm the ordering
        assert spans[0] is ot_scope.span._dd_span
        assert spans[1] is ot_scope2.span._dd_span
        assert spans[2] is dd_span
        assert spans[3] is ot_scope3.span._dd_span
        assert spans[4] is dd_span2

        # check the parenting
        assert spans[0].parent_id is None
        assert spans[1].parent_id == spans[0].span_id
        assert spans[2].parent_id == spans[1].span_id
        assert spans[3].parent_id == spans[2].span_id
        assert spans[4].parent_id == spans[3].span_id

    def test_consecutive_trace(self, ot_tracer, dd_tracer, test_spans):
        """Ensure consecutive usage of the opentracer and ddtracer."""
        with ot_tracer.start_active_span("my_ot_span") as ot_scope:
            pass

        with dd_tracer.trace("my_dd_span") as dd_span:
            pass

        with ot_tracer.start_active_span("my_ot_span") as ot_scope2:
            pass

        with dd_tracer.trace("my_dd_span") as dd_span2:
            pass

        spans = test_spans.pop()
        assert len(spans) == 4

        # confirm the ordering
        assert spans[0] is ot_scope.span._dd_span
        assert spans[1] is dd_span
        assert spans[2] is ot_scope2.span._dd_span
        assert spans[3] is dd_span2

        # check the parenting
        assert spans[0].parent_id is None
        assert spans[1].parent_id is None
        assert spans[2].parent_id is None
        assert spans[3].parent_id is None

    def test_ddtrace_wrapped_fn(self, ot_tracer, dd_tracer, test_spans):
        """Ensure ddtrace wrapped functions work with the opentracer"""

        @dd_tracer.wrap()
        def fn():
            with ot_tracer.start_span("ot_span_inner"):
                pass

        with ot_tracer.start_active_span("ot_span_outer"):
            fn()

        spans = test_spans.pop()
        assert len(spans) == 3

        # confirm the ordering
        assert spans[0].name == "ot_span_outer"
        assert spans[1].name == "tests.opentracer.core.test_dd_compatibility.fn"
        assert spans[2].name == "ot_span_inner"

        # check the parenting
        assert spans[0].parent_id is None
        assert spans[1].parent_id is spans[0].span_id
        assert spans[2].parent_id is spans[1].span_id

    def test_distributed_trace_propagation(self, ot_tracer, dd_tracer, test_spans):
        """Ensure that a propagated span context is properly activated."""
        span_ctx = SpanContext(trace_id=123, span_id=456)
        carrier = {}
        ot_tracer.inject(span_ctx, Format.HTTP_HEADERS, carrier)

        # extract should activate the span so that a subsequent start_span
        # will inherit from the propagated span context
        ot_tracer.extract(Format.HTTP_HEADERS, carrier)

        with dd_tracer.trace("test") as span:
            pass

        assert span.parent_id == 456
        assert span.trace_id == 123

        spans = test_spans.pop()
        assert len(spans) == 1
